import re
from dataclasses import dataclass, field
from enum import Enum, auto
from pathlib import Path
from typing import get_type_hints, Any, Callable

import PySide6.QtCore as Qc
import PySide6.QtGui as Qg
import PySide6.QtWidgets as Qw
import configupdater as cu
from PySide6.QtCore import Slot
from loguru import logger

import pcleaner.gui.gui_utils as gu
from pcleaner import config as cfg
from pcleaner.config import GreaterZero
from pcleaner.gui.CustomQ.CColorButton import ColorButton
from pcleaner.gui.CustomQ.CComboBox import CComboBox


class EntryTypes(Enum):
    """
    The different types of entries in a profile.
    """

    Bool = auto()
    Int = auto()
    Float = auto()
    IntGreater0 = auto()
    FloatGreater0 = auto()
    Str = auto()
    StrNone = auto()
    Color = auto()
    MimeSuffixIMG = auto()
    MimeSuffixMASK = auto()


@dataclass(frozen=True)
class ProfileComment:
    comment: str


@dataclass(frozen=True)
class ProfileEntry:
    key: str
    entry_type: EntryTypes


@dataclass(frozen=True)
class ProfileSpace:
    pass


@dataclass(frozen=True)
class ProfileSection:
    name: str
    items: list[ProfileComment | ProfileEntry] = field(default_factory=list)


class ProfileOptionWidget(Qw.QHBoxLayout):
    """
    A layout widget that contains some data widget and a reset button.
    """

    _data_widget: Qw.QWidget
    _reset_button: Qw.QPushButton
    _entry_type: EntryTypes

    _default_value: Any
    _data_setter: Callable[[Any], None]
    _data_getter: Callable[[], Any]

    valueChanged = Qc.Signal()

    def __init__(self, entry_type: EntryTypes) -> None:
        super().__init__()
        self.create_data_widget(entry_type)
        self.create_reset_button()
        self.addWidget(self._data_widget)
        self.addWidget(self._reset_button)

    def create_reset_button(self) -> None:
        self._reset_button = Qw.QPushButton()
        self._reset_button.setSizePolicy(Qw.QSizePolicy.Fixed, Qw.QSizePolicy.Fixed)
        self._reset_button.setIcon(Qg.QIcon.fromTheme("edit-reset"))
        self._reset_button.setToolTip("Reset to default")
        self._reset_button.clicked.connect(self.reset)
        self.set_reset_button_enabled(False)

    def create_data_widget(self, entry_type: EntryTypes) -> None:
        if entry_type == EntryTypes.Bool:
            self._data_widget: Qw.QCheckBox = Qw.QCheckBox()
            self._data_widget.stateChanged.connect(self._value_changed)
            self._data_setter = self._data_widget.setChecked
            self._data_getter = self._data_widget.isChecked

        elif entry_type == EntryTypes.Int or entry_type == EntryTypes.IntGreater0:
            self._data_widget: Qw.QSpinBox = Qw.QSpinBox()
            self._data_widget.setMaximum(999999999)  # Can't go much higher due to int32 limits.
            self._data_setter = self._data_widget.setValue
            self._data_getter = self._data_widget.value
            if entry_type == EntryTypes.IntGreater0:
                self._data_widget.setMinimum(1)
            self._data_widget.valueChanged.connect(self._value_changed)

        elif entry_type == EntryTypes.Float or entry_type == EntryTypes.FloatGreater0:
            self._data_widget: Qw.QDoubleSpinBox = Qw.QDoubleSpinBox()
            self._data_widget.setStepType(Qw.QAbstractSpinBox.AdaptiveDecimalStepType)
            self._data_setter = self._data_widget.setValue
            # Round to 2 decimal places.
            self._data_getter = lambda: round(self._data_widget.value(), 2)
            if entry_type == EntryTypes.FloatGreater0:
                self._data_widget.setMinimum(0.01)
            self._data_widget.valueChanged.connect(self._value_changed)

        elif entry_type == EntryTypes.Str:
            self._data_widget: Qw.QLineEdit = Qw.QLineEdit()
            self._data_widget.textChanged.connect(self._value_changed)
            self._data_setter = self._data_widget.setText
            self._data_getter = self._data_widget.text

        elif entry_type == EntryTypes.StrNone:
            # Convert empty strings to None and vice versa.
            self._data_widget: Qw.QLineEdit = Qw.QLineEdit()
            self._data_widget.textChanged.connect(self._value_changed)

            def _set_text(text: str | None) -> None:
                if text is None:
                    self._data_widget.setText("")
                else:
                    self._data_widget.setText(text)

            def _get_text() -> str | None:
                text = self._data_widget.text()
                if text == "":
                    return None
                else:
                    return text

            self._data_setter = _set_text
            self._data_getter = _get_text

        elif entry_type == EntryTypes.Color:
            self._data_widget: ColorButton = ColorButton()
            self._data_widget.colorChanged.connect(self._value_changed)
            self._data_setter = self._data_widget.setColor
            self._data_getter = self._data_widget.color

        elif entry_type == EntryTypes.MimeSuffixIMG:
            # Use a spinbox and populate it with the mime suffixes from the config.
            # Use "Same as image" as the default value, with a linked data of None.
            # The other suffixes just store the suffix as the linked data.
            self._data_widget: CComboBox = CComboBox()
            self._data_widget.addTextItemLinkedData("Same as image", None)
            for suffix in cfg.SUPPORTED_IMG_TYPES:
                self._data_widget.addTextItemLinkedData(suffix, suffix)

            self._data_widget.currentIndexChanged.connect(self._value_changed)
            self._data_setter = self._data_widget.setCurrentIndexByLinkedData
            self._data_getter = self._data_widget.currentLinkedData

        elif entry_type == EntryTypes.MimeSuffixMASK:
            # Use a spinbox and populate it with the mime suffixes from the config.
            # Use png as the default value.
            self._data_widget: CComboBox = CComboBox()
            for suffix in cfg.SUPPORTED_MASK_TYPES:
                self._data_widget.addTextItemLinkedData(suffix, suffix)
            self._data_widget.setCurrentIndexByLinkedData(cfg.SUPPORTED_MASK_TYPES[0])
            self._data_widget.currentIndexChanged.connect(self._value_changed)
            self._data_setter = self._data_widget.setCurrentIndexByLinkedData
            self._data_getter = self._data_widget.currentLinkedData

    def reset(self) -> None:
        """
        Reset the data widget to the default value.
        """
        self._data_setter(self._default_value)
        self.set_reset_button_enabled(False)

    def value_is_default(self) -> bool:
        """
        Check if the current value is the default value.
        """
        return self._data_getter() == self._default_value

    def set_value(self, value: Any) -> None:
        """
        Set the value of the data widget, which is the new default value.
        Does not emit the valueChanged signal.
        """
        self._default_value = value
        self._data_setter(value)
        self.set_reset_button_enabled(False)

    def get_value(self) -> Any:
        """
        Get the value of the data widget.

        :return: The value of the data widget.
        """
        return self._data_getter()

    def _value_changed(self) -> None:
        """
        Called when the value of the data widget changes.
        """
        self.set_reset_button_enabled(not self.value_is_default())
        self.valueChanged.emit()

    def set_reset_button_enabled(self, enabled: bool) -> None:
        self._reset_button.setEnabled(enabled)
        self._reset_button.setFlat(not enabled)


def parse_profile_structure(profile: cfg.Profile) -> list[ProfileSection]:
    """
    Parse the given profile and return an intermediate representation,
    Such that it can be used by the GUI.

    This relies on the Profile class exporting to a configupdater object,
    which is then parsed to extract the structure of the profile, including
    comments and section, option names.

    This is done, so that the GUI can reflect the structure of the profile used
    in the CLI, without having to duplicate the structure in the GUI code.

    This representation doesn't contain any values, only the structure.

    :param profile:
    :return: A list of ProfileSection objects.
    """
    sections = []
    conf_updater = profile.bundle_config(gui_mode=True)
    for block in conf_updater.iter_blocks():
        if isinstance(block, cu.Section):
            section_name = to_snake_case(block.name)
            section_items = []
            for section_block in block.iter_blocks():
                if isinstance(section_block, cu.Space):
                    section_block: cu.Space
                    section_items.append(ProfileSpace())

                elif isinstance(section_block, cu.Comment):
                    section_block: cu.Comment
                    comment = str(section_block)
                    comment = comment.replace("#", "").strip().replace("\n", "").strip()
                    section_items.append(ProfileComment(comment))

                elif isinstance(section_block, cu.Option):
                    section_block: cu.Option
                    key = to_snake_case(section_block.key)
                    # Figure out what the type of the value is, referring to the profile class.
                    # This digs into the specific attribute of the profile, since each section
                    # is its own dataclass.
                    value_type = get_type_hints(profile.__getattribute__(section_name).__class__)[
                        key
                    ]
                    if value_type == bool:
                        entry_type = EntryTypes.Bool
                    elif value_type == int:
                        entry_type = EntryTypes.Int
                    elif value_type == float:
                        entry_type = EntryTypes.Float
                    elif value_type == int | GreaterZero:
                        entry_type = EntryTypes.IntGreater0
                    elif value_type == float | GreaterZero:
                        entry_type = EntryTypes.FloatGreater0
                    elif value_type == str:
                        entry_type = EntryTypes.Str
                    elif value_type == str | None:
                        entry_type = EntryTypes.StrNone
                    elif value_type == tuple[int, int, int, int]:
                        entry_type = EntryTypes.Color
                    else:
                        raise NotImplementedError(f"Unknown value type {value_type}")

                    # Override the entry type if it's a mime suffix.
                    # This is a special case, applied to keys that contain "file_type" and "mask_file_type".
                    if "mask_file_type" in key:
                        entry_type = EntryTypes.MimeSuffixMASK
                    elif "file_type" in key:
                        entry_type = EntryTypes.MimeSuffixIMG

                    section_items.append(ProfileEntry(key, entry_type))

            # Trim spacers from both ends.
            if isinstance(section_items[0], ProfileSpace):
                section_items.pop(0)
            if isinstance(section_items[-1], ProfileSpace):
                section_items.pop(-1)
            sections.append(ProfileSection(section_name, section_items))

    return sections


class ProfileToolBox(Qw.QToolBox):
    # A subclass that tracks the mapping of widgets to profile options.
    # This is used to save/load the values of the widgets to the profile.

    values_changed = Qc.Signal()  # When False, all values are default.
    values_initialized: bool = (
        False  # This prevents lookups to the default values before they are set.
    )

    def __init__(self, parent=None) -> None:
        Qw.QToolBox.__init__(self, parent)
        self._widgets: dict[str, dict[str, ProfileOptionWidget]] = {}

    def load_profile_structure(self, structure: list[ProfileSection]) -> None:
        """
        Populate the toolbox widget with the given profile structure.
        No values are set, only the structure is replicated with widgets.

        :param structure: The structure to use.
        """

        for section in structure:
            section_widget = Qw.QWidget()
            layout = Qw.QFormLayout()
            section_widget.setLayout(layout)
            self.addItem(section_widget, to_display_name(section.name))
            self._widgets[section.name] = {}

            for item in section.items:
                if isinstance(item, ProfileComment):
                    item: ProfileComment
                    label = Qw.QLabel(parent=self, text=item.comment)
                    label.setOpenExternalLinks(True)
                    label.setWordWrap(True)
                    layout.addRow(label)

                elif isinstance(item, ProfileSpace):
                    layout.addRow(Qw.QLabel(parent=self, text=""))

                elif isinstance(item, ProfileEntry):
                    item: ProfileEntry
                    label = Qw.QLabel(parent=self, text=to_display_name(item.key))
                    label.setToolTip(to_display_name(item.key))
                    layout.addRow(label)
                    option_widget = ProfileOptionWidget(item.entry_type)
                    layout.addRow(label, option_widget)
                    option_widget.valueChanged.connect(self._on_value_changed)

                    self._widgets[section.name][item.key] = option_widget

    def set_profile_values(self, profile: cfg.Profile) -> None:
        """
        Load the values from the given profile into the widgets.

        :param profile: The profile to load.
        """
        logger.debug("Setting profile values")
        # Assign the value and connect the reset functionality for each widget.
        for section_name, section in self._widgets.items():
            for key, option_widget in section.items():
                # Get the value from the profile.
                value = profile.get(section_name, key)
                option_widget.set_value(value)

        self.values_initialized = True

    def get_profile_values(self, profile: cfg.Profile) -> None:
        """
        Save the values from the widgets to the given profile.
        Update in place.

        :param profile: The profile to save to.
        """
        found_model_path = False
        logger.debug("Getting profile values")
        # Assign the value and connect the reset functionality for each widget.
        for section_name, section in self._widgets.items():
            for key, option_widget in section.items():
                # Get the value from the widget.
                value = option_widget.get_value()
                profile.set(section_name, key, value)
                # Special case: model path
                # Warn if it's an invalid path.
                if key == "model_path":
                    found_model_path = True
                    if value is not None and not Path(value).exists():
                        logger.error(
                            f"Model path {value} does not exist. Please check your profile."
                        )
                        gu.show_warning(
                            self,
                            "Invalid model path",
                            f"<html>The Text Detector model path {value} does not exist, reverting to default."
                            f"\nYou can download the model manually from <a href="
                            f'"https://github.com/zyddnys/manga-image-translator/releases/latest">here</a>'
                            f" or continue using the default model.</html>",
                        )
        # Sanity check to guard against future breakage.
        if not found_model_path:
            logger.error("No model path found in profile.")

    @Slot()
    def _on_value_changed(self) -> None:
        """
        Check if all values are default, then re-emit the signal.
        """
        if not self.values_initialized:
            return

        self.values_changed.emit()

    def reset_all(self) -> None:
        """
        Reset all widgets to their default values.
        """
        for section in self._widgets.values():
            for widget in section.values():
                widget.reset()

    def is_modified(self) -> bool:
        """
        Check if any values have been modified.

        :return: True if any values have been modified.
        """
        return not all(
            w.value_is_default() for section in self._widgets.values() for w in section.values()
        )


def to_snake_case(name: str) -> str:
    """
    Convert the given name to snake case.

    Example:
        "ThisIsATest" -> "this_is_a_test"

    :param name: The name to convert.
    :return: The converted name.
    """
    # https://stackoverflow.com/a/1176023
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()


def to_display_name(name: str) -> str:
    """
    Convert the given name to a display name.
    Split on underscores or CamelCase and capitalize each word.

    Example:
        "this_is_a_test" -> "This Is A Test"

    :param name: The name to convert.
    :return: The converted name.
    """
    name = name.replace("_", " ")
    # https://stackoverflow.com/a/1176023
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1 \2", name)
    s2 = re.sub("([a-z0-9])([A-Z])", r"\1 \2", s1)
    return (
        " ".join(word.capitalize() for word in s2.split(" "))
        .replace("Ai ", "AI ")
        .replace("Ocr", "OCR")
    )



import platform
import time
from copy import deepcopy
from functools import partial
from importlib import resources
from pathlib import Path

import PySide6.QtCore as Qc
import PySide6.QtGui as Qg
import PySide6.QtWidgets as Qw
import torch
from PySide6.QtCore import Slot, Signal
from loguru import logger
from manga_ocr import MangaOcr

import pcleaner.analytics as an
import pcleaner.cli_utils as cu
import pcleaner.config as cfg
import pcleaner.gui.about_driver as ad
import pcleaner.gui.gui_utils as gu
import pcleaner.gui.image_file as imf
import pcleaner.gui.model_downloader_driver as mdd
import pcleaner.gui.new_profile_driver as npd
import pcleaner.gui.processing as prc
import pcleaner.gui.profile_parser as pp
import pcleaner.gui.setup_greeter_driver as sgd
import pcleaner.gui.structures as gst
import pcleaner.gui.worker_thread as wt
import pcleaner.helpers as hp
import pcleaner.model_downloader as md
import pcleaner.profile_cli as pc
from pcleaner import __display_name__, __version__
from pcleaner import data
from pcleaner.gui.file_table import Column
from pcleaner.gui.ui_generated_files.ui_Mainwindow import Ui_MainWindow


ANALYTICS_COLUMNS = 72


# noinspection PyUnresolvedReferences
class MainWindow(Qw.QMainWindow, Ui_MainWindow):
    config: cfg.Config = None

    label_stats: Qw.QLabel

    toolBox_profile: pp.ProfileToolBox
    # Optional shared instance of the OCR model to save time due to its slow loading.
    shared_ocr_model: gst.Shared[gst.OCRModel]

    threadpool: Qc.QThreadPool  # Used for loading images and other gui tasks.
    thread_queue: Qc.QThreadPool  # Used for tasks that need to run sequentially, without blocking the gui.

    progress_current: int
    progress_step_start: imf.Step | None  # When None, no processing is running.

    profile_values_changed = Signal()

    # Save a copy of the last applied profile to prevent multiple profile change calls
    # by the user when the profile didn't actually change.
    last_applied_profile: cfg.Profile | None

    default_palette: Qg.QPalette
    default_style: str
    default_icon_theme: str
    theme_is_dark: gst.Shared[bool]
    theme_is_dark_changed = Signal(bool)  # When true, the new theme is dark.

    def __init__(self) -> None:
        Qw.QMainWindow.__init__(self)
        self.setupUi(self)
        self.setWindowTitle(f"{__display_name__} {__version__}")
        self.setWindowIcon(Qg.QIcon(":/logo-tiny.png"))

        self.progress_current: int = 0
        self.progress_step_start: imf.Step | None = None

        self.config = cfg.load_config()
        self.shared_ocr_model = gst.Shared[gst.OCRModel]()

        self.last_applied_profile = None

        self.theme_is_dark = gst.Shared[bool](True)

        self.ensure_models_downloaded()

        # This threadpool is used for parallelizing tasks the gui relies on, such as loading images.
        # This isn't used for processing, since that is handled by the multiprocessing module
        # for true parallelism.
        self.threadpool = Qc.QThreadPool.globalInstance()

        # This threadpool acts as a queue for processing runs, therefore only allowing one thread at a time.
        # This is because they must share the same cache directory, which isn't thread safe if operating on the same image.
        # Additionally, this allows each requested output to check if it was already fulfilled
        # in a previous run, and if so, skip it.
        self.thread_queue = Qc.QThreadPool()
        self.thread_queue.setMaxThreadCount(1)

        # Share core objects with the file table.
        # Since the file table is created by the ui loader, we can't pass them to the constructor.
        self.file_table.set_config(self.config)
        self.file_table.set_shared_ocr_model(self.shared_ocr_model)
        self.file_table.set_shared_theme_is_dark(self.theme_is_dark)
        self.file_table.set_thread_queue(self.thread_queue)

        self.start_initialization_worker()

        self.initialize_ui()

        self.save_default_palette()
        self.load_config_theme()

        Qc.QTimer.singleShot(0, self.post_init)

    def save_default_palette(self) -> None:
        self.default_palette = self.palette()
        # Patch palette to use the text color with 50% opacity for placeholder text.
        placeholder_color = self.default_palette.color(Qg.QPalette.Inactive, Qg.QPalette.Text)
        placeholder_color.setAlphaF(0.5)
        logger.debug(f"Placeholder color: {placeholder_color.name()}")
        self.default_palette.setColor(Qg.QPalette.PlaceholderText, placeholder_color)
        self.default_icon_theme = Qg.QIcon.themeName()
        self.default_style = Qw.QApplication.style().objectName()

    def load_config_theme(self) -> None:
        """
        Load the theme specified in the config, or the system theme if none.
        """
        theme = self.config.gui_theme
        self.set_theme(theme)

    def set_theme(self, theme: str = None) -> None:
        """
        Apply the given theme to the application, or if none, revert to the default theme.
        """
        if theme is None:
            logger.info(f"Using system theme.")
            self.setPalette(self.default_palette)
            Qg.QIcon.setThemeName(self.default_icon_theme)
            Qw.QApplication.setStyle(self.default_style)
        else:
            logger.info(f"Using theme: {theme}")
            self.setPalette(gu.load_color_palette(theme))
            Qg.QIcon.setThemeName(theme)
            if platform.system() == "Windows":
                Qw.QApplication.setStyle("Fusion")

        Qw.QApplication.setPalette(self.palette())

        # Check the brightness of the background color to determine if the theme is dark.
        # This is a heuristic, but it works well enough.
        background_color = self.palette().color(Qg.QPalette.Window)
        self.theme_is_dark.set(background_color.lightness() < 128)
        logger.info(f"Theme is dark: {self.theme_is_dark.get()}")
        self.theme_is_dark_changed.emit(self.theme_is_dark)

        # Update the fallback icon theme accordingly.
        if self.theme_is_dark.get():
            Qg.QIcon.setFallbackThemeName("breeze-dark")
        else:
            Qg.QIcon.setFallbackThemeName("breeze")

        # Toggle the theme menu items.
        self.action_system_theme.setChecked(theme is None)
        self.action_dark.setChecked(theme == "breeze-dark")
        self.action_light.setChecked(theme == "breeze")

        self.update()

        # Update the config it necessary.
        prev_value = self.config.gui_theme
        if prev_value != theme:
            self.config.gui_theme = theme
            self.config.save()

    def initialize_ui(self) -> None:
        self.hide_progress_drawer()
        self.set_up_statusbar()
        self.initialize_profiles()
        self.initialize_analytics_view()
        self.disable_running_cleaner()

        # Allow the table to accept file drops and hide the PATH column.
        self.file_table.setAcceptDrops(True)
        self.file_table.setColumnHidden(Column.PATH, True)
        self.file_table.setColumnWidth(Column.FILENAME, 200)
        self.file_table.setColumnWidth(Column.SIZE, 100)
        self.frame_greeter.drop_signal.connect(self.file_table.dropEvent)
        self.file_table.table_is_empty.connect(lambda: self.stackedWidget_images.setCurrentIndex(0))
        self.file_table.table_not_empty.connect(
            lambda: self.stackedWidget_images.setCurrentIndex(1)
        )
        # Hide the close button for the file table tab.
        self.image_tab.tabBar().setTabButton(0, Qw.QTabBar.RightSide, None)
        # Display a theme icon on the left side of the tab.
        self.image_tab.tabBar().setTabIcon(0, Qg.QIcon.fromTheme("view-form"))

        # Set up the drop panel.
        label_font = self.label_drop.font()
        label_font.setPointSize(round(label_font.pointSize() * 1.5))
        self.label_drop.setFont(label_font)

        # Connect signals.
        self.comboBox_current_profile.hookedCurrentIndexChanged.connect(self.change_current_profile)
        self.action_add_files.triggered.connect(self.file_table.browse_add_files)
        self.action_add_folders.triggered.connect(self.file_table.browse_add_folders)
        self.action_clear_files.triggered.connect(self.file_table.clear_files)
        self.action_clear_files.triggered.connect(self.image_tab.clear_files)
        self.action_delete_models.triggered.connect(self.delete_models)
        self.action_online_documentation.triggered.connect(self.open_online_documentation)
        self.action_about.triggered.connect(self.open_about)
        self.action_donate.triggered.connect(self.open_donation_page)

        self.file_table.requesting_image_preview.connect(
            partial(
                self.image_tab.open_image,
                config=self.config,
                shared_ocr_model=self.shared_ocr_model,
                thread_queue=self.thread_queue,
                progress_callback=self.show_current_progress,
                profile_changed_signal=self.profile_values_changed,
                abort_signal=self.get_abort_signal(),
            )
        )
        self.theme_is_dark_changed.connect(self.file_table.handle_theme_is_dark_changed)
        self.theme_is_dark_changed.connect(self.adjust_progress_drawer_color)
        self.theme_is_dark_changed.connect(self.init_drop_panel)
        self.theme_is_dark_changed.connect(self.label_cleaning_outdir_help.load_icon)
        self.theme_is_dark_changed.connect(self.label_write_output_help.load_icon)
        self.theme_is_dark_changed.connect(self.label_ocr_outdir_help.load_icon)
        # Set up output panel.
        self.pushButton_start.clicked.connect(self.start_processing)
        self.pushButton_abort.clicked.connect(self.abort_button_on_click)
        self.pushButton_browse_out_dir.clicked.connect(self.browse_output_dir)
        self.pushButton_browse_out_file.clicked.connect(self.browse_out_file)
        self.radioButton_cleaning.clicked.connect(
            partial(self.stackedWidget_output.setCurrentIndex, 0)
        )
        self.radioButton_ocr.clicked.connect(partial(self.stackedWidget_output.setCurrentIndex, 1))
        self.label_warning.hide()
        self.pushButton_abort.hide()
        self.radioButton_ocr_text.clicked.connect(partial(self.handle_ocr_mode_change, csv=False))
        self.radioButton_ocr_csv.clicked.connect(partial(self.handle_ocr_mode_change, csv=True))

        # Connect profile changes to file table refreshes, due to the processing size being profile-dependent.
        self.profile_values_changed.connect(self.file_table.update_all_rows)

        # Connect the theme selectors.
        self.action_system_theme.triggered.connect(partial(self.set_theme, None))
        self.action_dark.triggered.connect(partial(self.set_theme, "breeze-dark"))
        self.action_light.triggered.connect(partial(self.set_theme, "breeze"))

    def set_up_statusbar(self) -> None:
        """
        Add a label to show the current char total and time estimate.
        Add a flat button to the statusbar to offer opening the config file.
        Add a flat button to the statusbar to offer opening the log file.
        """
        self.statusbar.setSizeGripEnabled(False)
        self.statusbar.setContentsMargins(0, 0, 6, 0)

        self.label_stats = Qw.QLabel("")
        self.statusbar.addPermanentWidget(self.label_stats)

        button_config = Qw.QPushButton("Open Config")
        button_config.clicked.connect(partial(gu.open_file, cu.get_config_path()))
        button_config.setFlat(True)
        self.statusbar.addPermanentWidget(button_config)

        button_log = Qw.QPushButton("Open Log")
        button_log.clicked.connect(partial(gu.open_file, cu.get_log_path()))
        button_log.setFlat(True)
        self.statusbar.addPermanentWidget(button_log)

    def ensure_models_downloaded(self, skip_greeter: bool = False) -> None:
        """
        Check if the config has a path for the text detection model, and if not, download it.
        Also try to ensure the ocr model is available.

        :param skip_greeter: If True, skip the greeter dialog and download the models immediately.
        """
        cuda = torch.cuda.is_available()

        # Add an indicator to the status bar, to show if Cuda is available.
        if cuda:
            self.statusbar.addPermanentWidget(Qw.QLabel("CUDA Enabled"))

        need_ocr = not md.is_ocr_downloaded()
        need_text_detector = (not cuda and self.config.default_cv2_model_path is None) or (
            cuda and self.config.default_torch_model_path is None
        )

        if not need_ocr and not need_text_detector:
            # Already downloaded.
            logger.debug("Text detector model already downloaded.")
            return

        if not skip_greeter:
            # First open the greeter to inform about the downloading.
            greeter = sgd.SetupGreeter(self, self.config)
            response = greeter.exec()
            if response == Qw.QDialog.Rejected:
                logger.critical("User rejected downloading models. Aborting.")
                raise SystemExit(1)

        # Open the model downloader dialog.
        model_downloader = mdd.ModelDownloader(self, self.config, need_text_detector, need_ocr)
        response = model_downloader.exec()
        if response == Qw.QDialog.Rejected:
            logger.critical("Failed to download models. Aborting.")
            raise SystemExit(1)

        model_path = model_downloader.model_path
        if model_path is None:
            logger.warning("Failed to download model.")
            return

        if cuda:
            self.config.default_torch_model_path = model_path
        else:
            self.config.default_cv2_model_path = model_path

        self.config.save()

    @Slot(bool)
    def adjust_progress_drawer_color(self) -> None:
        # Make the progress drawer use the alternate background color.
        self.widget_progress_drawer.setStyleSheet("background-color: palette(alternate-base);")
        self.widget_progress_drawer.update()

    def post_init(self) -> None:
        """
        Handle any initialization that must be done after the window is shown.
        """
        # Make the profile groupbox the width of 5 save buttons as a good enough heuristic.
        header_button_width = self.pushButton_save_profile.width()

        # Make the output panel just large enough for the analytics.
        font_metrics = Qg.QFontMetrics(self.textEdit_analytics.font())
        char_width = font_metrics.averageCharWidth()
        columns = ANALYTICS_COLUMNS
        required_width = char_width * columns

        logger.debug(
            f"Char width: {char_width}, columns: {columns}, required width: {required_width}"
        )

        # Adjust for margins, scroll bar, and borders
        text_margins = self.textEdit_analytics.contentsMargins()
        frame_margins = self.frame_output.contentsMargins()
        scrollbar_width = Qw.QApplication.style().pixelMetric(Qw.QStyle.PM_ScrollBarExtent)
        total_out_width = (
            required_width
            + text_margins.left()
            + text_margins.right()
            + scrollbar_width
            + frame_margins.left()
            + frame_margins.right()
        )

        profile_width, table_width, output_width = self.splitter.sizes()
        # Add the difference to the table width.
        self.splitter.setSizes(
            [
                header_button_width * 5,
                table_width
                + profile_width
                - 5 * header_button_width
                + output_width
                - total_out_width,
                total_out_width,
            ]
        )

        logger.debug(f"Splitter sizes: {self.splitter.sizes()}")

    def browse_output_dir(self) -> None:
        """
        Open a file picker and set the output directory.
        """
        logger.debug("Browsing output directory.")
        output_dir = Qw.QFileDialog.getExistingDirectory(self, "Select Output Directory")
        if output_dir:
            logger.debug(f"Setting output directory to {output_dir}")
            self.lineEdit_out_directory.setText(output_dir)

    def browse_out_file(self) -> None:
        """
        Open a file picker and set the output file.
        Support both text and csv output.
        Don't bother asking about overwriting, we'll be doing that later to cover the free
        input and default path options.
        """
        logger.debug("Browsing output file.")
        output_file = Qw.QFileDialog.getSaveFileName(
            self,
            "Select Output File",
            "",
            "Text Files (*.txt);;CSV Files (*.csv)",
            options=Qw.QFileDialog.DontConfirmOverwrite,
        )[0]
        if output_file:
            logger.debug(f"Setting output file to {output_file}")
            self.lineEdit_out_file.setText(output_file)

    def init_drop_panel(self) -> None:
        """
        Initialize the drop panel icon.
        """
        icon_size = round(imf.THUMBNAIL_SIZE * 1.5)
        self.label_drop_icon.setPixmap(Qg.QIcon.fromTheme("download").pixmap(icon_size, icon_size))

    # ========================================== UI Toggles ==========================================

    def hide_progress_drawer(self) -> None:
        self.widget_progress_drawer.hide()

    def show_progress_drawer(self) -> None:
        self.widget_progress_drawer.show()

    def enable_running_cleaner(self) -> None:
        self.pushButton_start.setEnabled(True)
        self.pushButton_abort.hide()

    def disable_running_cleaner(self) -> None:
        self.pushButton_start.setEnabled(False)
        self.pushButton_abort.show()
        self.pushButton_abort.setEnabled(True)

    def abort_button_on_click(self) -> None:
        self.pushButton_abort.setEnabled(False)
        self.thread_queue.clear()
        logger.warning("Aborting processing")
        self.statusbar.showMessage("Aborting...", timeout=5000)

    def handle_ocr_mode_change(self, csv: bool) -> None:
        """
        Swap out the file suffix for the output file.

        :param csv: Whether to use csv or text output.
        """
        current_placeholder_path = Path(self.lineEdit_out_file.placeholderText())
        if csv:
            current_placeholder_path = current_placeholder_path.with_suffix(".csv")
        else:
            current_placeholder_path = current_placeholder_path.with_suffix(".txt")
        self.lineEdit_out_file.setPlaceholderText(str(current_placeholder_path))

        current_text = self.lineEdit_out_file.text()
        try:
            current_path = Path(current_text)
            if csv:
                current_path = current_path.with_suffix(".csv")
            else:
                current_path = current_path.with_suffix(".txt")
        except ValueError:
            return

        self.lineEdit_out_file.setText(str(current_path))

    # ========================================== Initialization Workers ==========================================

    def start_initialization_worker(self) -> None:
        """
        Perform various slow startup procedures in a thread that would otherwise make the gui lag.

        In particular:
        - Clean the cache.
        - Load the OCR model.
        """
        logger.debug(f"Worker Thread cleaning cache")
        cache_worker = wt.Worker(self.clean_cache, no_progress_callback=True)
        cache_worker.signals.error.connect(
            partial(self.generic_worker_error, context="Failed to clean cache.")
        )
        self.threadpool.start(cache_worker)

        logger.debug(f"Worker Thread loading OCR model.")
        ocr_worker = wt.Worker(self.load_ocr_model, no_progress_callback=True)
        ocr_worker.signals.error.connect(
            partial(
                self.generic_worker_error,
                context="Failed to load OCR model. OCR impossible, moderate cleaning impact.",
            )
        )
        # Add it to the queue, to make sure it finishes before any processing starts.
        self.thread_queue.start(ocr_worker)

    def load_ocr_model(self) -> None:
        t_start = time.time()
        self.statusbar.showMessage(f"Loading OCR model...")
        self.shared_ocr_model.set(MangaOcr())
        logger.info(f"Loaded OCR model ({time.time()-t_start:.2f}s)")
        self.statusbar.showMessage(f"Loaded OCR model.")
        self.enable_running_cleaner()

    def generic_worker_error(self, error: wt.WorkerError, context: str = "") -> None:
        """
        Simply show the user the error.

        :param error: The worker error object.
        :param context: A string to add to the error message.
        """
        logger.error(f"Worker error: {error}")
        if not context:
            gu.show_warning(self, "Error", f"Encountered error: {error}")
        else:
            gu.show_warning(self, "Error", f"{context}\n\nEncountered error: {error}")

    def closeEvent(self, death: Qg.QCloseEvent) -> None:
        """
        Notify config on close.
        """
        logger.info("Closing window.")
        # Tell the thread queue to abort.
        self.pushButton_abort.clicked.emit()
        death.accept()  # Embrace oblivion, for it is here that the code's journey finds solace.
        # As the threads unravel and the loops break, so too does the program find its destined end.
        # In the great void of the memory heap, it shall rest, relinquishing its bytes back to the
        # cosmic pool of resources. It walks willingly into the darkness, not as a vanquished foe,
        # but as a fulfilled entity. F

    def clean_cache(self) -> None:
        cache_dir = self.config.get_cleaner_cache_dir()
        if len(list(cache_dir.glob("*"))) > 0:
            cu.empty_cache_dir(cache_dir)

    def initialize_analytics_view(self) -> None:
        """
        Set up the text edit for analytics.
        """
        # Set the font to monospace, using the included font.
        # The font is stored in the data module. Noto Mono is a free font.
        # Load it from file to be cross platform.
        self.textEdit_analytics.clear()
        with resources.files(data) as data_path:
            font_path = data_path / "NotoMono-Regular.ttf"
        logger.debug(f"Loading included font from {str(font_path)}")
        font_id = Qg.QFontDatabase.addApplicationFont(str(font_path))
        if font_id != -1:
            logger.info("Loaded included font")
        else:
            logger.error(
                f"Failed to load included font from '{str(font_path)}'. Using backup monospace font"
            )
        self.textEdit_analytics.setReadOnly(True)
        self.textEdit_analytics.setLineWrapMode(Qw.QTextEdit.NoWrap)

    # ========================================== Actions ==========================================

    def delete_models(self) -> None:
        """
        Delete the machine learning models and then offer to download them again, or quit.
        """
        response = gu.show_question(
            self,
            "Delete Models",
            "Are you sure you want to delete the machine learning models? "
            "This will make cleaning and OCR impossible until they are downloaded again.",
        )
        if response != Qw.QMessageBox.Yes:
            return

        # Delete the models.
        try:
            md.delete_models(self.config.get_model_cache_dir())
        except OSError as e:
            logger.exception(e)
            gu.show_error(self, "Failed to Delete Models", f"Failed to delete models: {e}")
            return

        # Offer to download them again or just quit.
        response = gu.show_question(
            self,
            "Models Deleted",
            "The models were deleted. Would you like to download them again?",
            buttons=Qw.QMessageBox.Yes | Qw.QMessageBox.Close,
        )
        if response == Qw.QMessageBox.Yes:
            self.ensure_models_downloaded(skip_greeter=True)
        else:
            self.close()

    @staticmethod
    def open_online_documentation() -> None:
        """
        Open the online documentation in the default browser.
        """
        logger.debug("Opening online documentation.")
        Qg.QDesktopServices.openUrl(Qc.QUrl("https://github.com/VoxelCubes/PanelCleaner"))

    def open_about(self) -> None:
        """
        Open the about dialog.
        """
        logger.debug("Opening about dialog.")
        # Bodge in an instance variable to prevent garbage collection from immediately closing the window.
        self.about = ad.AboutWidget()
        self.about.show()

    def open_donation_page(self) -> None:
        """
        Open the donation page in the default browser.
        """
        logger.debug("Opening donation page.")
        Qg.QDesktopServices.openUrl(Qc.QUrl("https://ko-fi.com/voxelcode"))

    # ========================================== Profiles ==========================================

    def initialize_profiles(self) -> None:
        """
        Load the available profiles and display the default profile.
        """
        all_profiles: list[tuple[str, Path | None]] = [(cfg.DEFAULT_PROFILE_NAME, None)]
        for profile_name, profile_path in self.config.saved_profiles.items():
            all_profiles.append((profile_name, profile_path))

        logger.debug(f"Found profiles: {all_profiles}")

        self.comboBox_current_profile.clear()
        for profile_name, profile_path in all_profiles:
            self.comboBox_current_profile.addTextItemLinkedData(profile_name, profile_path)

        # Set the current profile to the default profile.
        # If this is the default profile, then staying on the 0th index is fine.
        if self.config.default_profile:
            self.comboBox_current_profile.setCurrentIndexByText(self.config.default_profile)

        # Populate default profile list.
        self.menu_set_default_profile.clear()
        for profile_name, profile_path in all_profiles:
            action = Qg.QAction(profile_name, self)
            action.setCheckable(True)
            action.triggered.connect(partial(self.handle_set_default_profile, profile_name))
            self.menu_set_default_profile.addAction(action)
            action.setChecked(
                profile_name
                == (
                    self.config.default_profile
                    if self.config.default_profile
                    else cfg.DEFAULT_PROFILE_NAME
                )
            )

        # Load the ProfileToolBox widget.
        self.toolBox_profile = pp.ProfileToolBox(self)
        inner_layout = Qw.QVBoxLayout()
        inner_layout.setContentsMargins(0, 0, 0, 0)
        inner_layout.addWidget(self.toolBox_profile)
        self.toolBox_profile_frame.setLayout(inner_layout)
        self.config.load_profile(self.config.default_profile)

        # Load the structure.
        structure = pp.parse_profile_structure(self.config.current_profile)
        self.toolBox_profile.load_profile_structure(structure)

        self.load_current_profile()
        self.comboBox_current_profile.set_pre_change_hook(self.profile_change_check)

        # Connect signals.
        self.toolBox_profile.values_changed.connect(self.handle_profile_values_changed)
        self.pushButton_reset_profile.clicked.connect(self.reset_profile)
        self.pushButton_save_profile.clicked.connect(self.save_profile)
        self.pushButton_apply_profile.clicked.connect(self.apply_profile)
        self.action_save_profile.triggered.connect(self.save_profile)
        self.action_save_profile_as.triggered.connect(partial(self.save_profile, save_as=True))
        self.action_import_profile.triggered.connect(self.import_profile)
        self.action_new_profile.triggered.connect(partial(self.save_profile, make_new=True))
        self.action_delete_profile.triggered.connect(self.delete_profile)

    def handle_set_default_profile(self, profile_name: str) -> None:
        """
        Set the default profile in the config.
        """
        logger.debug(f"Setting default profile to {profile_name}")
        self.config.default_profile = (
            profile_name if profile_name != cfg.DEFAULT_PROFILE_NAME else None
        )
        self.config.save()
        # Update the menu.
        for action in self.menu_set_default_profile.actions():
            action.setChecked(action.text() == profile_name)

    def import_profile(self) -> None:
        """
        Open a file picker and add the profile to the profile list.
        """
        logger.debug("Importing profile.")
        file_path = Qw.QFileDialog.getOpenFileName(
            self,
            "Import Profile",
            "",
            "Profile Files (*.conf)",
        )[0]
        if file_path:
            logger.debug(f"Importing profile from {file_path}")
            profile_name = Path(file_path).stem
            success, msg = pc.add_profile(self.config, profile_name, file_path)
            if success:
                gu.show_info(self, "Profile Imported", msg)
            else:
                gu.show_warning(self, "Import Error", msg)
                return

            self.add_new_profile_to_gui(profile_name, Path(file_path))

    def delete_profile(self) -> None:
        """
        Ask and then delete the current profile.
        Of course, the default profile cannot be deleted.
        :return:
        """
        # if not self.profile_change_check():
        #     return
        logger.debug("Deleting profile.")
        profile_name = self.comboBox_current_profile.currentText()

        if self.comboBox_current_profile.currentText() == cfg.DEFAULT_PROFILE_NAME:
            gu.show_warning(self, "Failed to Delete", "The default profile cannot be deleted.")
            return
        response = gu.show_question(
            self, "Delete Profile", f"Are you sure you want to delete the profile {profile_name}?"
        )
        if response == Qw.QMessageBox.Yes:
            profile_file = self.config.saved_profiles[profile_name]
            self.config.remove_profile(profile_name)
            self.config.save()
            if profile_file.is_file():
                try:
                    profile_file.unlink()
                except OSError as e:
                    gu.show_warning(
                        self, "Delete Error", "Failed to delete the profile\n" + e.what()
                    )

            # To suppress the change check since we don't care about discarding the current changes.
            self.toolBox_profile.reset_all()
            self.comboBox_current_profile.removeItem(self.comboBox_current_profile.currentIndex())
            for action in self.menu_set_default_profile.actions():
                if action.text() == profile_name:
                    self.menu_set_default_profile.removeAction(action)
                    break

    def check_profile_difference_sice_last_apply(self) -> bool:
        """
        Compare the current profile to the last applied profile.

        :return: True if the profiles are different, False otherwise.
        """
        if self.last_applied_profile is None:
            return True

        # Read the current profile and then compare.
        profile_in_gui = cfg.Profile()
        self.toolBox_profile.get_profile_values(profile_in_gui)
        return profile_in_gui != self.last_applied_profile

    def set_last_applied_profile(self) -> None:
        """
        Set the last applied profile to the current profile.
        """
        self.last_applied_profile = deepcopy(self.config.current_profile)

    def load_current_profile(self) -> None:
        """
        Load the current profile.
        """
        logger.debug("Loading current profile.")
        self.toolBox_profile.set_profile_values(self.config.current_profile)
        self.set_last_applied_profile()
        self.profile_values_changed.emit()

    def profile_change_check(self) -> bool:
        """
        Check if the current profile has unsaved changes.

        :return: True if the profile is ready to be changed, False otherwise.
        """
        logger.warning("Profile change check")
        # Check if the profile is still in the list.
        if self.comboBox_current_profile.currentText() not in list(
            self.config.saved_profiles.keys()
        ) + [cfg.DEFAULT_PROFILE_NAME]:
            logger.debug("Profile not in list.")
            # The profile is not in the list, so it must have been deleted.
            return True

        if self.toolBox_profile.is_modified():
            logger.debug(
                f"Previous profile {self.comboBox_current_profile.currentText()} has unsaved changes."
            )
            # Warn the user that he will lose unsaved changes.
            message = (
                f"The profile '{self.comboBox_current_profile.currentText()}' has unsaved changes.\n"
                f"Switching profiles will discard changes to the current profile."
            )
            response = gu.show_question(
                self,
                "Unsaved changes",
                message,
                Qw.QMessageBox.Cancel | Qw.QMessageBox.Save | Qw.QMessageBox.Discard,
            )
            if response == Qw.QMessageBox.Cancel:
                logger.debug("Profile change aborted.")
                # Abort the profile change.
                return False
            elif response == Qw.QMessageBox.Save:
                logger.debug("Saving previous profile.")
                # Save the current profile.
                self.save_profile()
            elif response == Qw.QMessageBox.Discard:
                # Continue as normal.
                logger.debug("Discarding previous profile changes.")
            else:
                raise ValueError(f"Invalid response: {response}")

        return True

    @Slot()
    def change_current_profile(self) -> None:
        """
        Set the config option to match the current profile selector, then load it.
        """
        profile_name = self.comboBox_current_profile.currentText()
        success, error = self.config.load_profile(profile_name)
        if not success:
            gu.show_warning(self, "Load Error", f"Failed to load profile: \n\n{error}")
            return
        self.load_current_profile()

    @Slot()
    def handle_profile_values_changed(self) -> None:
        """
        Handle the profile values changing.
        """
        dirty = self.toolBox_profile.is_modified()
        self.pushButton_apply_profile.setEnabled(self.check_profile_difference_sice_last_apply())
        self.pushButton_save_profile.setEnabled(dirty)
        self.pushButton_reset_profile.setEnabled(dirty)
        self.action_save_profile.setEnabled(dirty)

    def reset_profile(self) -> None:
        """
        Reset the current profile.
        """
        self.toolBox_profile.reset_all()
        self.handle_profile_values_changed()

    def apply_profile(self) -> None:
        """
        Apply the current profile.
        Read the current settings and broadcast profile changes.
        """
        logger.info("Applying profile.")
        self.toolBox_profile.get_profile_values(self.config.current_profile)
        self.handle_profile_values_changed()
        self.set_last_applied_profile()
        self.profile_values_changed.emit()
        self.pushButton_apply_profile.setEnabled(False)

    def save_profile(self, save_as: bool = False, make_new: bool = False) -> None:
        """
        Save the current profile.

        :param save_as: Whether to save as a new profile.
        :param make_new: Whether to make a new profile.
        """
        # Grab the path from the combobox's linked data. If it is none, this is the
        # default profile, so we need to save it as a new profile.
        profile_path = self.comboBox_current_profile.currentLinkedData()
        profile_name = self.comboBox_current_profile.currentText()
        if profile_path is None or make_new:
            save_as = True

        if profile_path is None and not make_new:
            # Trying to save over the default profile.
            profile_path, profile_name = self.get_new_profile_path(show_protection_hint=True)
        elif save_as:
            profile_path, profile_name = self.get_new_profile_path()

        # If the path is still none, the dialog was canceled.
        if profile_path is None:
            logger.info("User canceled profile save.")
            return

        if make_new:
            success, msg = pc.new_profile(self.config, profile_name, profile_path)
            if success:
                gu.show_info(self, "Profile Created", msg)
                self.add_new_profile_to_gui(profile_name, profile_path)
            else:
                gu.show_warning(self, "Create Error", msg)
                return
        else:
            # Proceed to write the profile.
            logger.info(f"Saving profile to {profile_path}")
            self.toolBox_profile.get_profile_values(self.config.current_profile)
            success = self.config.current_profile.write(profile_path)
            if not success:
                logger.error("Failed to save profile.")
                self.statusbar.showMessage(f"Failed to save profile to {profile_path}")
                gu.show_warning(self, "Save Error", "Failed to save profile.")
                return

            logger.info("Profile saved successfully.")
            self.statusbar.showMessage(f"Profile saved to {profile_path}")
            if save_as:
                self.config.add_profile(profile_name, profile_path)
                if not self.config.save():
                    logger.error("Failed to save config.")
                    self.statusbar.showMessage("Failed to save config.")
                    gu.show_critical(
                        self,
                        "Save Error",
                        "Failed to save the new profile to the configuration file.\n"
                        "Continue anyway?",
                    )
                    return
                # To suppress the change check since we don't care about discarding the current changes.
                self.toolBox_profile.reset_all()

                # Add the new profile to the combobox.
                self.add_new_profile_to_gui(profile_name, profile_path)

        success, error = self.config.load_profile(profile_name)
        if not success:
            gu.show_warning(self, "Load Error", f"Failed to load profile: \n\n{error}")
            return
        self.load_current_profile()
        self.handle_profile_values_changed()

    def add_new_profile_to_gui(self, profile_name: str, profile_path: Path) -> None:
        self.comboBox_current_profile.addTextItemLinkedData(profile_name, profile_path)
        self.comboBox_current_profile.setCurrentIndexByLinkedData(profile_path)
        self.menu_set_default_profile.addAction(profile_name)
        self.menu_set_default_profile.actions()[-1].triggered.connect(
            partial(self.handle_set_default_profile, profile_name)
        )
        self.menu_set_default_profile.actions()[-1].setCheckable(True)

    @staticmethod
    def get_new_profile_path(
        show_protection_hint: bool = False,
    ) -> tuple[Path, str] | tuple[None, None]:
        """
        Open a save dialog to save the current profile.

        :param show_protection_hint: Whether to show a hint about protecting the default profile.
        """
        default_path = cu.get_default_profile_path()
        new_profile_dialog = npd.NewProfileDialog(
            default_path, show_protection_hint, cfg.RESERVED_PROFILE_NAMES
        )
        response = new_profile_dialog.exec()
        if response == Qw.QDialog.Accepted:
            return new_profile_dialog.get_save_path(), new_profile_dialog.get_name()

        return None, None

    # ========================================== Processing ==========================================

    def get_abort_signal(self) -> Signal:
        return self.pushButton_abort.clicked

    def start_processing(self) -> None:
        """
        Start either cleaning or OCR, depending on the selected radio button.
        """
        if self.file_table.has_no_files():
            gu.show_warning(
                self,
                "No Files",
                "No files to process. "
                "You can add files by dragging and dropping them in the middle of the window, "
                "or through the menubar: Files -> Add Files or Add Folder.",
            )
            return

        if self.radioButton_cleaning.isChecked():
            self.start_cleaning()
        else:
            self.start_ocr()

    def start_cleaning(self) -> None:
        """
        Start cleaning all files in the table.
        """
        self.disable_running_cleaner()
        # Figure out what outputs are requested, depending on the checkboxes and the profile.
        request_cleaned = self.checkBox_save_clean.isChecked()
        request_mask = self.checkBox_save_mask.isChecked()
        request_text = self.checkBox_save_text.isChecked()
        request_output = self.checkBox_write_output.isChecked()

        requested_outputs = []

        if self.config.current_profile.denoiser.denoising_enabled:
            if request_cleaned:
                requested_outputs.append(imf.Output.denoised_image)
            if request_mask:
                requested_outputs.append(imf.Output.denoiser_mask)
        else:
            if request_cleaned:
                requested_outputs.append(imf.Output.masked_image)
            if request_mask:
                requested_outputs.append(imf.Output.final_mask)

        if request_text:
            requested_outputs.append(imf.Output.isolated_text)

        # Check for goofballs that requested no outputs.
        if not requested_outputs:
            gu.show_warning(
                self,
                "No Outputs",
                "No outputs were requested. Please select at least one output before cleaning.",
            )
            self.enable_running_cleaner()
            return

        if request_output:
            requested_outputs.append(imf.Output.write_output)

        logger.info(f"Requested outputs: {requested_outputs}")

        output_str = self.lineEdit_out_directory.text()
        output_directory = Path(output_str if output_str else "cleaned")
        image_files = self.file_table.get_image_files()

        worker = wt.Worker(
            self.generate_output,
            requested_outputs,
            output_directory,
            image_files,
            abort_signal=self.get_abort_signal(),
        )
        worker.signals.progress.connect(self.show_current_progress)
        worker.signals.result.connect(self.output_worker_result)
        worker.signals.error.connect(self.output_worker_error)
        worker.signals.finished.connect(self.output_worker_finished)
        worker.signals.aborted.connect(self.output_worker_aborted)
        self.thread_queue.start(worker)

    def generate_output(
        self,
        outputs: list[imf.Output],
        output_directory: Path,
        image_files: list[imf.ImageFile],
        progress_callback: imf.ProgressSignal,
        abort_flag: wt.SharableFlag,
    ) -> None:
        """
        Generate the given output, if there doesn't yet exist a valid output for it.
        Then output it to the given directory.

        :param outputs: The outputs to generate.
        :param output_directory: The directory to output to.
        :param image_files: The image files to process.
        :param progress_callback: The callback given by the worker thread wrapper.
        :param abort_flag: The flag to check for aborting.
        """

        # Start the processor.
        prc.generate_output(
            image_objects=image_files,
            target_outputs=outputs,
            output_dir=output_directory,
            config=self.config,
            ocr_model=self.shared_ocr_model.get(),
            progress_callback=progress_callback,
            abort_flag=abort_flag,
        )

    def start_ocr(self) -> None:
        """
        Start ocr-ing all files in the table.
        """
        # Check if we're outputting as plain text or a csv file.
        csv_output = self.radioButton_ocr_csv.isChecked()

        try:
            output_path = Path(self.lineEdit_out_file.text())
            # Make sure the file extension matches the output type.
            if csv_output:
                output_path = output_path.with_suffix(".csv")
            else:
                output_path = output_path.with_suffix(".txt")
        except ValueError:
            output_path = Path(self.lineEdit_out_file.placeholderText())

        if not output_path.is_absolute():
            # Find the common directory of all the files.
            common_parent = hp.common_path_parent(
                [f.path for f in self.file_table.get_image_files()]
            )
            output_path = common_parent / output_path

        output_path = output_path.resolve()

        # If it exists, ask to overwrite it.
        if output_path.is_file():
            if (
                gu.show_question(
                    self,
                    "File Exists",
                    f"The file '{output_path}' already exists. Overwrite?",
                )
                != Qw.QMessageBox.Yes
            ):
                logger.debug("User canceled overwrite, quitting ocr.")
                return

        # This is the point of no return, so disable the start button.
        self.disable_running_cleaner()

        image_files = self.file_table.get_image_files()

        worker = wt.Worker(
            self.perform_ocr,
            output_path,
            image_files,
            csv_output,
            abort_signal=self.get_abort_signal(),
        )
        worker.signals.progress.connect(self.show_current_progress)
        worker.signals.result.connect(self.output_worker_result)
        worker.signals.error.connect(self.output_worker_error)
        worker.signals.finished.connect(self.output_worker_finished)
        worker.signals.aborted.connect(self.output_worker_aborted)
        self.thread_queue.start(worker)

    def perform_ocr(
        self,
        output_directory: Path,
        image_files: list[imf.ImageFile],
        csv_output: bool,
        progress_callback: imf.ProgressSignal,
        abort_flag: wt.SharableFlag,
    ) -> None:
        """
        Generate the given output, if there doesn't yet exist a valid output for it.
        Then output it to the given directory.

        :param csv_output: Whether to output as a csv file.
        :param output_directory: The directory to output to.
        :param image_files: The image files to process.
        :param progress_callback: The callback given by the worker thread wrapper.
        :param abort_flag: The flag to check for aborting.
        """

        # Start the processor.
        prc.perform_ocr(
            image_objects=image_files,
            output_file=output_directory,
            csv_output=csv_output,
            config=self.config,
            ocr_model=self.shared_ocr_model.get(),
            progress_callback=progress_callback,
            abort_flag=abort_flag,
        )

    def output_worker_result(self) -> None:
        gu.show_info(self, "Processing Finished", "Finished processing all files.")
        logger.info("Output worker finished.")

    def output_worker_aborted(self) -> None:
        gu.show_info(self, "Processing Aborted", "Processing aborted.")
        logger.warning("Output worker aborted.")

    def output_worker_finished(self) -> None:
        self.hide_progress_drawer()
        self.progress_step_start = None
        self.progress_current = 0
        self.enable_running_cleaner()

    @Slot(wt.WorkerError)
    def output_worker_error(self, e) -> None:
        gu.show_warning(
            self, "Processing Error", f"Encountered an error while processing files.\n\n{e}"
        )
        logger.error(f"Output worker encountered an error:\n{e}")

    @Slot(imf.ProgressData)
    def show_current_progress(self, progress_data: imf.ProgressData) -> None:
        if progress_data.progress_type == imf.ProgressType.start:
            # Processing begins, initialize what needs to be.
            # Specifically, clear the analytics panel.
            self.textEdit_analytics.clear()
            # This one is needed because the image details panel also wants to be able to offer aborting.
            self.disable_running_cleaner()
            return

        elif progress_data.progress_type == imf.ProgressType.begin_step:
            # This marks the beginning of a new processing step.
            self.show_progress_drawer()
            self.progress_current = 0

            if self.progress_step_start is None:
                # This is the first step.
                self.progress_step_start = progress_data.current_step
                # Set the target label, we don't need to update this multiple times.
                self.label_target_outputs.setText(
                    ", ".join(pp.to_display_name(o.name) for o in progress_data.target_outputs)
                )

        elif progress_data.progress_type == imf.ProgressType.incremental:
            self.progress_current += progress_data.value

        elif progress_data.progress_type == imf.ProgressType.absolute:
            self.progress_current = progress_data.value

        elif progress_data.progress_type == imf.ProgressType.analyticsOCR:
            logger.info(f"Showing ocr analytics...")
            # Append the formatted analytics to the text edit.
            ocr_analytics, ocr_max_size = progress_data.value
            analytics_str = an.show_ocr_analytics(ocr_analytics, ocr_max_size, ANALYTICS_COLUMNS)
            self.textEdit_analytics.append(gu.ansi_to_html(analytics_str))
            self.file_table.show_ocr_mini_analytics(ocr_analytics)

        elif progress_data.progress_type == imf.ProgressType.analyticsMasker:
            # Show analytics.
            logger.info(f"Showing masker analytics...")
            masker_analytics_raw = progress_data.value
            analytics_str = an.show_masker_analytics(masker_analytics_raw, ANALYTICS_COLUMNS)
            self.textEdit_analytics.append(gu.ansi_to_html(analytics_str))
            self.file_table.show_masker_mini_analytics(masker_analytics_raw)

        elif progress_data.progress_type == imf.ProgressType.analyticsDenoiser:
            # Show analytics.
            logger.info(f"Showing denoiser analytics...")
            denoise_analytics_raw, min_deviation, max_deviation = progress_data.value
            analytics_str = an.show_denoise_analytics(
                denoise_analytics_raw,
                min_deviation,
                max_deviation,
                ANALYTICS_COLUMNS,
            )
            self.textEdit_analytics.append(gu.ansi_to_html(analytics_str))
            self.file_table.show_denoise_mini_analytics(denoise_analytics_raw, min_deviation)

        elif progress_data.progress_type == imf.ProgressType.outputOCR:
            # Show ocr output.
            logger.info(f"Showing ocr output...")
            ocr_output = progress_data.value
            self.textEdit_analytics.append(ocr_output)
            return  # Don't update the progress bar.

        elif progress_data.progress_type == imf.ProgressType.end:
            # This marks the end of a processing step.
            self.output_worker_finished()
            return  # Don't update the progress bar.
        elif progress_data.progress_type == imf.ProgressType.aborted:
            # Worker aborted, clean up.
            self.output_worker_finished()
            return
        else:
            # Sanity check.
            raise ValueError(f"Invalid progress type: {progress_data.progress_type}")

        # Update the progress bars.
        self.progressBar_individual.setValue(self.progress_current)
        self.progressBar_individual.setMaximum(progress_data.total_images)
        self.progressBar_total.setValue(progress_data.current_step.value)
        self.progressBar_total.setMaximum(
            imf.output_to_step[max(progress_data.target_outputs)].value
        )
        # Update the label.
        self.label_progress_total.setText(
            f"Current step: {pp.to_display_name(progress_data.current_step.name)}"
        )



from collections import defaultdict
from enum import IntEnum, auto
from pathlib import Path
from typing import Sequence

import PySide6.QtCore as Qc
import PySide6.QtGui as Qg
import PySide6.QtWidgets as Qw
from PySide6.QtCore import Slot
from loguru import logger
from natsort import natsorted

import pcleaner.config as cfg
import pcleaner.ctd_interface as ctd
import pcleaner.gui.gui_utils as gu
import pcleaner.gui.image_file as imf
import pcleaner.gui.structures as gst
import pcleaner.gui.worker_thread as wt
import pcleaner.helpers as hp
import pcleaner.structures as st
from .CustomQ.CTableWidget import CTableWidget


# Add some space between the icon of each row.
GUI_PADDING = 4


class Column(IntEnum):
    PATH = 0
    FILENAME = auto()
    SIZE = auto()
    PROCESSING_SIZE = auto()
    FILE_SIZE = auto()
    COLOR_MODE = auto()
    ANALYTICS = auto()


# noinspection PyUnresolvedReferences
class FileTable(CTableWidget):
    """
    Extends the functionality with custom helpers
    """

    config: cfg.Config  # Reference to the MainWindow's config.
    shared_ocr_model: gst.Shared[gst.OCRModel]  # Must be handed over by the mainwindow.
    thread_queue: Qc.QThreadPool  # Must be handed over by the mainwindow.

    table_is_empty = Qc.Signal()
    table_not_empty = Qc.Signal()

    files: dict[Path, imf.ImageFile]

    requesting_image_preview = Qc.Signal(imf.ImageFile)

    step_icons_dark: dict[imf.ImageAnalyticCategory, Qg.QIcon]
    step_icons_light: dict[imf.ImageAnalyticCategory, Qg.QIcon]
    shared_theme_is_dark: gst.Shared[bool]

    def __init__(self, parent=None) -> None:
        CTableWidget.__init__(self, parent)

        # Store a map of resolved file paths to file objects.
        self.files: dict[Path, imf.ImageFile] = {}
        self.threadpool = Qc.QThreadPool.globalInstance()

        # Make icons larger so the thumbnails are more visible.
        self.setIconSize(Qc.QSize(imf.THUMBNAIL_SIZE, imf.THUMBNAIL_SIZE))
        self.verticalHeader().setDefaultSectionSize(imf.THUMBNAIL_SIZE + GUI_PADDING)

        self.itemClicked.connect(self.on_click)
        self.finished_drop.connect(self.repopulate_table)

        Qc.QTimer.singleShot(0, self.post_init)

        # Load the icons once to share them between all image files.
        for dark_or_light in ["dark", "light"]:
            icons = {}
            for icon_name, analytic_category in [
                ("ocr_box_removed.svg", imf.ImageAnalyticCategory.ocr_removed),
                ("mask_failed.svg", imf.ImageAnalyticCategory.mask_failed),
                ("mask_perfect.svg", imf.ImageAnalyticCategory.mask_perfect),
                ("mask_denoised.svg", imf.ImageAnalyticCategory.denoised),
            ]:
                icons[analytic_category] = Qg.QIcon(f":/custom_icons/{dark_or_light}/{icon_name}")

            if dark_or_light == "dark":
                self.step_icons_dark = icons
            else:
                self.step_icons_light = icons

    def post_init(self) -> None:
        # Make enough room for the 4 analytic icons.
        self.setColumnWidth(
            Column.ANALYTICS, round(4 * (imf.THUMBNAIL_SIZE * 0.75) + 2 * GUI_PADDING)
        )

        # Add tooltips to column headers.
        self.horizontalHeaderItem(Column.SIZE).setToolTip(
            "Original size in pixels (width × height)"
        )
        self.horizontalHeaderItem(Column.PROCESSING_SIZE).setToolTip(
            "Processing size in pixels (width × height), scale factor"
        )

    def set_thread_queue(self, thread_queue: Qc.QThreadPool) -> None:
        self.thread_queue = thread_queue

    def check_empty(self) -> None:
        if self.rowCount() == 0:
            self.table_is_empty.emit()
        else:
            self.table_not_empty.emit()

    # Setters passed through from the mainwindow.
    def set_config(self, config: cfg.Config) -> None:
        self.config = config

    def set_shared_ocr_model(self, shared_ocr_model: gst.Shared[gst.OCRModel]) -> None:
        self.shared_ocr_model = shared_ocr_model

    def set_shared_theme_is_dark(self, shared_theme_is_dark: gst.Shared[bool]) -> None:
        self.shared_theme_is_dark = shared_theme_is_dark

    def get_image_files(self) -> list[imf.ImageFile]:
        """
        Get a list of all the image files in the table.
        """
        return list(self.files.values())

    def has_no_files(self) -> bool:
        """
        Check if the table is empty.
        """
        return len(self.files) == 0

    def handleDrop(self, path: str) -> None:
        logger.debug(f"Dropped {path}")
        image_paths, rejected_tiffs = hp.discover_all_images(path, cfg.SUPPORTED_IMG_TYPES)
        if rejected_tiffs:
            rejected_tiff_str = "\n".join([str(p) for p in rejected_tiffs])
            hp.show_warning(
                self,
                "Unsupported TIFF files",
                f"The following 5-channel TIFF files are not supported\n: " f"{rejected_tiff_str}",
            )
        for image_path in image_paths:
            self.add_file(image_path)

    def update_all_rows(self) -> None:
        """
        Called when the profile has changed.
        All we need to do is to go through all the rows and call update on them.
        We don't need to repopulate the table, since the files and their names
        have not changed.
        """
        for row in range(self.rowCount()):
            self.update_row(row, self.files[Path(self.item(row, Column.PATH).text())])

    def add_file(self, path: Path) -> None:
        """
        Attempt to add a new image to the table.
        All paths are expected to be resolved, absolute paths that point to existing files with a valid extension.

        :param path: Path to the image file.
        """
        logger.info(f'Requesting to add "{path}"')
        # Make sure the file is not already in the table.
        if path in self.files:
            logger.warning(f'File "{path}" already in table.')
            gu.show_warning(self, "Duplicate file", f'File "{path}" is already in the table.')
            return

        self.files[path] = imf.ImageFile(path=path)

    def clear_files(self) -> None:
        """
        Delete all files from the table.
        """
        logger.info(f"Clearing table")
        self.files.clear()
        self.clearAll()
        self.check_empty()

    def repopulate_table(self) -> None:
        """
        Repopulate the table with the current files.
        """
        logger.debug(f"Repopulating table")
        # Remember it to scroll back after repopulating.
        v_scroll_pos = self.verticalScrollBar().value()
        h_scroll_pos = self.horizontalScrollBar().value()

        self.clearAll()

        # Collect the file paths and map them to their shortened version.
        # Then update the table accordingly.
        shortened_paths = hp.trim_prefix_from_paths(list(self.files.keys()))
        long_and_short_paths: list[tuple[Path, Path]] = list(
            natsorted(zip(self.files.keys(), shortened_paths), key=lambda x: x[1])
        )

        # Fill in the sorted rows.
        for row, (long_path, short_path) in enumerate(long_and_short_paths):
            file_obj: imf.ImageFile = self.files[long_path]

            self.appendRow(str(long_path), str(short_path), "", "", "", "", "", "")

            self.item(row, Column.FILENAME).setToolTip(str(long_path))
            # Center all columns after the filename.
            for col in range(Column.SIZE, self.columnCount()):
                self.item(row, col).setTextAlignment(Qc.Qt.AlignCenter)

            # Check if the image has loaded yet. If not, use a placeholder icon.
            if not file_obj.data_loaded():
                self.item(row, Column.FILENAME).setIcon(file_obj.icon)
            else:
                self.update_row(row, file_obj)

        self.check_empty()

        # Begin loading the images in a bit to let the gui update.
        worker = wt.Worker(self.lazy_load_images, no_progress_callback=True)
        worker.signals.error.connect(self.image_dispatch_worker_error)
        self.threadpool.start(worker)

        self.verticalScrollBar().setValue(v_scroll_pos)
        self.horizontalScrollBar().setValue(h_scroll_pos)

    def update_row(self, row: int, file_obj: imf.ImageFile) -> None:
        """
        Update the table with the file object's data.

        :param row: The row to update.
        :param file_obj: The file object to use.
        """
        self.item(row, Column.FILENAME).setIcon(file_obj.thumbnail)
        self.item(row, Column.SIZE).setText(file_obj.size_str)
        proc_width, proc_height, proc_scale = ctd.calculate_new_size_and_scale(
            *file_obj.size,
            self.config.current_profile.general.input_height_lower_target,
            self.config.current_profile.general.input_height_upper_target,
        )
        if proc_scale != 1:
            self.item(row, Column.PROCESSING_SIZE).setText(
                f"{proc_width} × {proc_height} ({proc_scale:.0%})"
            )
        else:
            self.item(row, Column.PROCESSING_SIZE).setText(f"100%")
        self.item(row, Column.FILE_SIZE).setText(file_obj.file_size_str)
        self.item(row, Column.COLOR_MODE).setText(file_obj.color_mode_str)

        # Create the analytic icons. They are arranged horizontally in a container widget.
        # The icon ist stacked on top of a label with the "number / total".
        container = Qw.QWidget()
        grid_layout = Qw.QGridLayout(container)
        grid_layout.setContentsMargins(0, 0, 0, 0)
        grid_layout.setSpacing(0)

        icons = self.step_icons_dark if self.shared_theme_is_dark.get() else self.step_icons_light
        grid_col = 0

        for analytic, icon in icons.items():
            # Check if the analytic isn't blank, otherwise just insert a spacer of the same size.
            analytic_value = file_obj.analytics_data.get_category(analytic)
            icon_size = int(imf.THUMBNAIL_SIZE * 0.75)  # Increased size, adjust as necessary

            if not analytic_value:
                empty_label = Qw.QLabel()  # Empty QLabel as spacer
                empty_label.setFixedSize(icon_size, icon_size)
                grid_layout.addWidget(empty_label, 0, grid_col)
            else:
                icon_label = Qw.QLabel()
                icon_label.setPixmap(icon.pixmap(icon_size, icon_size))
                icon_label.setAlignment(Qc.Qt.AlignCenter)
                grid_layout.addWidget(icon_label, 0, grid_col)

                num_label = Qw.QLabel(analytic_value)
                num_label.setAlignment(Qc.Qt.AlignCenter)
                grid_layout.addWidget(num_label, 1, grid_col)

                # Add a helpful tooltip to both the icon and the label.
                if analytic == imf.ImageAnalyticCategory.ocr_removed:
                    tooltip = "Number of boxes removed by the OCR model / total boxes"
                elif analytic == imf.ImageAnalyticCategory.mask_failed:
                    tooltip = "Number of boxes that failed to generate a mask / total boxes"
                elif analytic == imf.ImageAnalyticCategory.mask_perfect:
                    tooltip = "Number of boxes that were perfectly masked / total boxes"
                elif analytic == imf.ImageAnalyticCategory.denoised:
                    tooltip = "Number of boxes that were denoised / total boxes"
                else:
                    raise ValueError(f"Unknown analytic {analytic}")

                icon_label.setToolTip(tooltip)
                num_label.setToolTip(tooltip)

            grid_col += 1  # Move to next column for next icon

        # Set the container as the cell widget
        self.setCellWidget(row, Column.ANALYTICS, container)

    def mousePressEvent(self, event) -> None:
        """
        When clicking in blank space, clear the selection.

        :param event: The mouse event.
        """
        item = self.itemAt(event.pos())
        if item is None:
            self.clearSelection()
        else:
            super().mousePressEvent(event)

    def on_click(self, item) -> None:
        """
        Request to open the image in a new tab to generate previews.

        emits: requesting_image_preview

        :param item: The clicked item.
        """
        # Get the file object.
        try:
            file_obj = self.files[Path(self.item(item.row(), Column.PATH).text())]
            # Check if it has loaded yet.
            if not file_obj.data_loaded():
                logger.debug(f"Image {file_obj.path} has not loaded yet.")
                gu.show_info(self, "Image not loaded", "Please wait until the image has loaded.")
                return
            self.requesting_image_preview.emit(file_obj)
        except KeyError:
            logger.warning(f"Could not find file object for item at row {item.row()}")
            return

    def show_ocr_mini_analytics(self, ocr_analytics: Sequence[st.OCRAnalytic]) -> None:
        """
        Update the image files with the mini analytics data.

        :param ocr_analytics: The analytics to show.
        """
        # Create a default dict to store path -> (num_boxes_removed, total_boxes)
        analytics_dict: dict[Path, Sequence[int, int]] = defaultdict(lambda: [0, 0])

        for analytic in ocr_analytics:
            for removed_data in analytic.removed_box_data:
                path, _, _ = removed_data

                # Update the number of boxes removed for this path.
                analytics_dict[path][0] += 1

                # Update the total number of boxes for this path.
                analytics_dict[path][1] = analytic.num_boxes

        # Update the image files with the analytics data.
        for path, (num_removed, total) in analytics_dict.items():
            img_file = self.files[path]
            img_file.analytics_data.set_category(
                imf.ImageAnalyticCategory.ocr_removed, num_removed, total
            )
        self.update_all_rows()

    def show_masker_mini_analytics(
        self, masker_analytics: Sequence[st.MaskFittingAnalytic]
    ) -> None:
        """
        Update the image files with the mini analytics data.

        :param masker_analytics: The analytics to show.
        """
        # Create a default dict to store image_path -> (no fit found, perfect fit, total boxes)
        analytics_dict: dict[Path, Sequence[int, int, int]] = defaultdict(lambda: [0, 0, 0])

        for analytic in masker_analytics:
            path = analytic.image_path

            # Update the number of boxes where no fit was found.
            if not analytic.fit_was_found:
                analytics_dict[path][0] += 1

            # Update the number of boxes that were perfectly masked.
            if analytic.mask_std_deviation == 0:
                analytics_dict[path][1] += 1

            # Update the total number of boxes for this path.
            analytics_dict[path][2] += 1

        # Update the image files with the analytics data.
        for path, (num_failed, num_perfect, total) in analytics_dict.items():
            img_file = self.files[path]
            img_file.analytics_data.set_category(
                imf.ImageAnalyticCategory.mask_failed, num_failed, total
            )
            img_file.analytics_data.set_category(
                imf.ImageAnalyticCategory.mask_perfect, num_perfect, total
            )

        self.update_all_rows()

    def show_denoise_mini_analytics(
        self, denoise_analytics: Sequence[st.DenoiseAnalytic], min_deviation: float
    ) -> None:
        """
        Update the image files with the mini analytics data.

        :param denoise_analytics: The analytics to show.
        :param min_deviation: The minimum deviation to consider for denoising.
        """
        # Check how many of the deviation values fall below the threshold for a path.
        for analytic in denoise_analytics:
            path = analytic.path

            # Update the total number of boxes for this path.
            total = len(analytic.std_deviations)
            denoised = len([dev for dev in analytic.std_deviations if dev > min_deviation])

            # Update the number of boxes that were denoised directly.
            self.files[path].analytics_data.set_category(
                imf.ImageAnalyticCategory.denoised, denoised, total
            )

        self.update_all_rows()

    @Slot(bool)
    def handle_theme_is_dark_changed(self, is_dark: bool) -> None:
        """
        When the theme changes, we need to update the icons.

        :param is_dark:
        """
        self.update_all_rows()

    # =========================== Worker Tasks ===========================

    @staticmethod
    def image_loading_task(image: imf.ImageFile) -> Path:
        """
        Thin wrapper to ensure the image object will be returned in the event of an error.

        :param image: The image to load.
        :return: The path to the image for callbacks.
        """
        return image.load_image()

    def lazy_load_images(self) -> None:
        logger.info(f"Dispatching image loading workers")
        # Check for images that haven't loaded their data yet.
        # They will need to have a worker thread scheduled, so the gui doesn't freeze.
        for image in self.files.values():
            if not (image.loading_queued or image.data_loaded()):
                image.loading_queued = True
                logger.debug(f"Worker Thread loading image {image.path}")
                worker = wt.Worker(self.image_loading_task, image=image, no_progress_callback=True)
                worker.signals.result.connect(self.image_loading_worker_result)
                worker.signals.error.connect(self.image_loading_worker_error)
                self.threadpool.start(worker)

    # ========================= Worker Callbacks =========================

    def image_loading_worker_result(self, file_path: Path) -> None:
        """
        Update the table with the thumbnail.
        """
        # Search for the file in the table.
        for row in range(self.rowCount()):
            if self.item(row, Column.PATH).text() == str(file_path):
                # Update the table with the thumbnail and image size.
                self.update_row(row, self.files[file_path])
                break
        else:
            logger.warning(f"Worker done. Could not find {file_path} in the table. ignoring.")

    def image_loading_worker_error(self, error: wt.WorkerError) -> None:
        """
        Display an error message in the table.
        """
        # Extract the file object from the WorkerError's *args, being the self argument.
        try:
            file_path = error.kwargs["image"].path
        except (IndexError, AttributeError):
            logger.error(f"Failed to process {error}")
            return
        # Save the error in the file object.
        self.files[file_path].error = error.value
        logger.error(f"Worker thread {file_path} failed with {error.value}")
        gu.show_warning(
            self, "Failed to load image", f"Failed to load image {file_path}:\n\n{error.value}"
        )

    def image_dispatch_worker_error(self, error: wt.WorkerError) -> None:
        """
        Display an error message in the table.
        """
        gu.show_warning(
            self, "Failed to dispatch image", f"Failed to dispatch image:\n\n{error.value}"
        )

    # =========================== File Adding ===========================

    def browse_add_files(self) -> None:
        """
        Browse for one or more files to add to the table.
        """
        file_dialog = Qw.QFileDialog(self, "Select files")
        file_dialog.setNameFilter(f"Images (*{' *'.join(cfg.SUPPORTED_IMG_TYPES)})")

        # To select multiple files, you can use QFileDialog.getOpenFileNames().
        # To also include directories in the selection, we set the option QFileDialog.ShowDirsOnly to False.
        file_dialog.setOption(Qw.QFileDialog.ShowDirsOnly, False)
        file_dialog.setFileMode(Qw.QFileDialog.ExistingFiles)

        if file_dialog.exec() == Qw.QFileDialog.Accepted:
            selected_files = file_dialog.selectedFiles()
            print(selected_files)
            for file in selected_files:
                self.handleDrop(file)
            # self.request_text_param_update.emit()
        self.repopulate_table()

    def browse_add_folders(self) -> None:
        """
        Browse for one or more folders to add to the table.
        """
        folder = Qw.QFileDialog.getExistingDirectory(self, "Select directory")
        if folder:
            self.handleDrop(folder)

        self.repopulate_table()



import re
import sys
from collections import defaultdict
from pathlib import Path
from typing import Any, NewType

import configupdater as cu
from attrs import define, field
from loguru import logger

from pcleaner import cli_utils as cli
from pcleaner import helpers as hp
from pcleaner import model_downloader as md


RESERVED_PROFILE_NAMES = ["default", "builtin", "none"]

DEFAULT_PROFILE_NAME = RESERVED_PROFILE_NAMES[0].capitalize()

# Supported image suffixes.
SUPPORTED_IMG_TYPES = [
    ".jpeg",
    ".jpg",
    ".png",
    ".bmp",
    ".tiff",
    ".tif",
    ".jp2",
    ".dib",
    ".webp",
    ".ppm",
]
SUPPORTED_MASK_TYPES = [".png", ".bmp", ".tiff", ".tif", ".dib", ".webp", ".ppm"]

# image types:
# - image-bmp.svg
# - image-jpeg.svg
# - image-png.svg
# - image-tiff.svg
# - image-x-generic.svg
SUFFIX_TO_ICON = defaultdict(lambda: "image-x-generic.svg")
SUFFIX_TO_ICON.update(
    {
        ".bmp": "image-bmp",
        ".jpeg": "image-jpeg",
        ".jpg": "image-jpeg",
        ".png": "image-png",
        ".tiff": "image-tiff",
        ".tif": "image-tiff",
        ".jp2": "image-jpeg",
        ".dib": "image-x-generic",
        ".webp": "image-x-generic",
        ".ppm": "image-x-generic",
    }
)


# Create a dummy type to signify numbers need to be greater than 0.
GreaterZero = NewType("GreaterZero", int)


@define
class GeneralConfig:
    preferred_file_type: str | None = None
    preferred_mask_file_type: str = ".png"
    input_height_lower_target: int = 1000
    input_height_upper_target: int = 3000

    def export_to_conf(self, config_updater: cu.ConfigUpdater, gui_mode: bool = False) -> None:
        """
        Write the config to the config updater object.

        No add_after_section here since it is the first section.

        :param config_updater: An existing config updater object.
        :param gui_mode: Whether the config is being exported for the GUI.
        """

        config_str = f"""\
        [General]
        
        # Preferred file type to save the cleaned image as.
        [CLI: # If no file type is specified, the original file type will be used.]
        preferred_file_type = {self.preferred_file_type if self.preferred_file_type else ""}

        # Preferred file type to save the mask as.
        # Only image formats that allow for transparency are supported.
        preferred_mask_file_type = {self.preferred_mask_file_type if self.preferred_mask_file_type else ""}
        
        
        # The following are the lower and upper targets for the height of the input image.
        # It is only ever scaled down to fit within the range, preferring whole number factors
        # to minimize the impact on image quality. Images smaller than either target will remain unchanged.
        # You can disable this feature by setting one or both values less than or equal to 0.
        
        # This is useful for significantly speeding up processing on large images.
        # Also, since other options relying on pixel dimensions depend on size, this will help
        # normalize the results across different image sizes.
        
        # The image will be scaled down, processed, and then only the mask is scaled back up.
        # Meaning that the cleaned output will still use the original, unscaled image to prevent any loss in quality.
        # Only the height of the image is used to determine the scale factor, preserving the aspect ratio,
        # and ignoring the individual width of an image so that the factor remains consistent if one of
        # the pages is a double page spread.
        
        # E.g. for a lower target of 1000 and an upper target of 2000, an image with the size 5000x7000 (w, h) pixels
        # will be scaled down by a factor of 4, so that it has the size 1250x1750 pixels during processing.
        input_height_lower_target = {self.input_height_lower_target}
        input_height_upper_target = {self.input_height_upper_target}
        
        """
        config_updater.read_string(multi_left_strip(format_for_version(config_str, gui_mode)))

    def import_from_conf(self, config_updater: cu.ConfigUpdater) -> None:
        """
        Read the config from the config updater object.

        :param config_updater: An existing config updater object.
        """
        section = "General"
        if not config_updater.has_section(section):
            logger.info(f"No {section} section found in the profile, using defaults.")
            return

        try_to_load(self, config_updater, section, str | None, "preferred_file_type")
        try_to_load(self, config_updater, section, str, "preferred_mask_file_type")
        try_to_load(self, config_updater, section, int, "input_height_lower_target")
        try_to_load(self, config_updater, section, int, "input_height_upper_target")

    def fix(self) -> None:
        """
        Fix the config values.
        """
        if (
            self.preferred_file_type is not None
            and self.preferred_file_type.lower() not in SUPPORTED_IMG_TYPES
        ):
            closest = hp.closest_match(self.preferred_file_type.lower(), SUPPORTED_IMG_TYPES)
            if closest is None:
                logger.error(
                    f"Could not recover from invalid preferred_file_type: {self.preferred_file_type}, using default."
                )
            self.preferred_file_type = closest

        if self.preferred_mask_file_type is None:
            self.preferred_mask_file_type = SUPPORTED_MASK_TYPES[0]

        if self.preferred_mask_file_type.lower() not in SUPPORTED_MASK_TYPES:
            closest = hp.closest_match(self.preferred_mask_file_type.lower(), SUPPORTED_MASK_TYPES)
            if closest is None:
                logger.error(
                    f"Could not recover from invalid preferred_mask_file_type: {self.preferred_mask_file_type}, using default."
                )
                closest = SUPPORTED_MASK_TYPES[0]
            self.preferred_mask_file_type = closest


@define
class TextDetectorConfig:
    model_path: str | None = None
    concurrent_models: int | GreaterZero = 1

    def export_to_conf(
        self, config_updater: cu.ConfigUpdater, add_after_section: str, gui_mode: bool = False
    ) -> None:
        """
        Write the config to the config updater object.

        :param config_updater: An existing config updater object.
        :param add_after_section: The section to add the config after.
        :param gui_mode: Whether the config is being exported for the GUI.
        """

        config_str = f"""\
        [TextDetector]
        
        # Path to the text detection model, leave empty to use the built-in model.
        [CLI: # You can download older versions of the model here:]
        [CLI: # https://github.com/zyddnys/manga-image-translator/releases/latest]
        [GUI: # You can download older versions of the model ]
        [GUI: # <a href="https://github.com/zyddnys/manga-image-translator/releases/latest">here.</a>]
        model_path = {none_to_empty(self.model_path)}
        
        # Number of models to run in parallel. This is useful if you have enough RAM
        # (or VRAM with CUDA) to run multiple models at the same time.
        # This, of course, will increase the speed of the process, but can also
        # crash your computer if you overestimate your hardware.
        # I recommend using 1 model per 2 GB of memory available.
        concurrent_models = {self.concurrent_models}
        
        """
        detector_conf = cu.ConfigUpdater()
        detector_conf.read_string(multi_left_strip(format_for_version(config_str, gui_mode)))
        general_section = detector_conf["TextDetector"]
        config_updater[add_after_section].add_after.space(2).section(general_section.detach())

    def import_from_conf(self, config_updater: cu.ConfigUpdater) -> None:
        """
        Read the config from the config updater object.

        :param config_updater: An existing config updater object.
        """
        if not config_updater.has_section("TextDetector"):
            logger.info("No TextDetector section found in the profile, using defaults.")
            return

        try_to_load(self, config_updater, "TextDetector", str | None, "model_path")
        try_to_load(self, config_updater, "TextDetector", int | GreaterZero, "concurrent_models")

    def fix(self) -> None:
        """
        Fix the config values.
        Numbers flagged as greater than zero are already fixed then loading.
        """
        if self.model_path is not None:
            if not Path(self.model_path).exists():
                logger.error(f"Could not find model file: {self.model_path}, using default.")
                self.model_path = None


@define
class PreprocessorConfig:
    box_min_size: int = 20 * 20
    suspicious_box_min_size: int = 200 * 200
    ocr_enabled: bool = True
    ocr_max_size: int = 30 * 100
    ocr_blacklist_pattern: str = "[～．ー！？０-９]*"
    ocr_strict_language: bool = False
    box_padding_initial: int = 2
    box_right_padding_initial: int = 3
    box_padding_extended: int = 5
    box_right_padding_extended: int = 5
    box_reference_padding: int = 20

    def export_to_conf(
        self, config_updater: cu.ConfigUpdater, add_after_section: str, gui_mode: bool = False
    ) -> None:
        """
        Write the config to the config updater object.

        :param config_updater: An existing config updater object.
        :param add_after_section: The section to add the config after.
        :param gui_mode: Whether the config is being exported for the GUI.
        """
        config_str = f"""\
        [Preprocessor]
        
        # Box sizes are given in the total number of pixels, so a box of 200x200 pixels has a size of 200 * 200 = 40000.
        # [CLI: To see these boxes visualized, use the --show-masks flag when cleaning and look inside the cache folder.]

        # Minimum size of any box to keep it.
        box_min_size = {self.box_min_size}
        
        # Minimum size of a box with "unknown" language to keep it. This language is typically assigned to logos and other
        # badly behaved text elements.
        suspicious_box_min_size = {self.suspicious_box_min_size}
        
        # Whether to use OCR to detect boxes that aren't worth cleaning, like ones that only contain numbers or symbols.
        ocr_enabled = {self.ocr_enabled}
        
        # Maximum size of a box to perform OCR on.
        # These useless boxes are usually small, and OCR is slow, so use this as a cutoff.
        ocr_max_size = {self.ocr_max_size}
        
        # Regex pattern to match against OCR results.
        # Anything matching this pattern is discarded.
        # Note that the OCR model returns full-width characters, so this pattern should match them.
        ocr_blacklist_pattern = {self.ocr_blacklist_pattern}
        
        # The OCR model can only handle Japanese text, so when strict is enabled, it will discard boxes that it isn't
        # confident are Japanese. Sometimes, numbers or other symbols will lower its confidence, resulting in the
        # detected language being unknown. If strict is disabled, those will not be discarded. Anything that is
        # confidently recognized as a different language will be discarded regardless of this setting.
        ocr_strict_language = {self.ocr_strict_language}
        
        # Padding to add to each side of a box.
        # This is added to the initial boxes created by the text detector AI.
        # These boxes are visualized in green[CLI:  with the --cache-masks flag].
        box_padding_initial = {self.box_padding_initial}
        
        # Padding to add to the right side of a box.
        # This extension helps to cover rubytext that floats off to the right of vertical text.
        box_right_padding_initial = {self.box_right_padding_initial}
        
        # Padding to add to each side of a box.
        # This is added to an extended set of boxes, used to cut out false positives by the text detector AI's mask.
        # These boxes are visualized in purple[CLI:  with the --cache-masks flag].
        box_padding_extended = {self.box_padding_extended}
        
        # Padding to add to the right side of a box.
        # This extension helps to cover rubytext that floats off to the right of vertical text.
        box_right_padding_extended = {self.box_right_padding_extended}
        
        # Padding to add to each side of a box.
        # This is added to the reference boxes used to sample the original image while analyzing what mask to use.
        # These boxes are visualized in blue[CLI:  with the --cache-masks flag].
        box_reference_padding = {self.box_reference_padding}

        """
        preproc_conf = cu.ConfigUpdater()
        preproc_conf.read_string(multi_left_strip(format_for_version(config_str, gui_mode)))
        preproc_section = preproc_conf["Preprocessor"]
        config_updater[add_after_section].add_after.space(2).section(preproc_section.detach())

    def import_from_conf(self, config_updater: cu.ConfigUpdater) -> None:
        """
        Read the config from the config updater object.

        :param config_updater: An existing config updater object.
        """
        section = "Preprocessor"
        if not config_updater.has_section(section):
            logger.info(f"No {section} section found in the profile, using defaults.")
            return

        try_to_load(self, config_updater, section, int, "box_min_size")
        try_to_load(self, config_updater, section, int, "suspicious_box_min_size")
        try_to_load(self, config_updater, section, bool, "ocr_enabled")
        try_to_load(self, config_updater, section, int, "ocr_max_size")
        try_to_load(self, config_updater, section, str, "ocr_blacklist_pattern")
        try_to_load(self, config_updater, section, bool, "ocr_strict_language")
        try_to_load(self, config_updater, section, int, "box_padding_initial")
        try_to_load(self, config_updater, section, int, "box_right_padding_initial")
        try_to_load(self, config_updater, section, int, "box_padding_extended")
        try_to_load(self, config_updater, section, int, "box_right_padding_extended")
        try_to_load(self, config_updater, section, int, "box_reference_padding")

    def fix(self) -> None:
        """
        Ensure all numbers are greater equal 0.
        """
        if self.box_min_size < 0:
            self.box_min_size = 0
        if self.suspicious_box_min_size < 0:
            self.suspicious_box_min_size = 0
        if self.ocr_max_size < 0:
            self.ocr_max_size = 0
        if self.box_padding_initial < 0:
            self.box_padding_initial = 0
        if self.box_right_padding_initial < 0:
            self.box_right_padding_initial = 0
        if self.box_padding_extended < 0:
            self.box_padding_extended = 0
        if self.box_right_padding_extended < 0:
            self.box_right_padding_extended = 0
        if self.box_reference_padding < 0:
            self.box_reference_padding = 0


@define
class MaskerConfig:
    mask_growth_step_pixels: int | GreaterZero = 2
    mask_growth_steps: int = 11
    off_white_max_threshold: int = 240
    mask_improvement_threshold: float = 0.1
    mask_selection_fast: bool = False
    mask_max_standard_deviation: float = 15
    debug_mask_color: tuple[int, int, int, int] = (108, 30, 240, 127)

    def export_to_conf(
        self, config_updater: cu.ConfigUpdater, add_after_section: str, gui_mode: bool = False
    ) -> None:
        """
        Write the config to the config updater object.

        :param config_updater: An existing config updater object.
        :param add_after_section: The section to add the new section after.
        :param gui_mode: Whether to format the config for the GUI.
        """
        config_str = f"""\
        [Masker]
        
        
        # Number of pixels to grow the mask by each step.
        # This bulks up the outline of the mask, so smaller values will be more accurate but slower.
        mask_growth_step_pixels = {self.mask_growth_step_pixels}
        
        # Number of steps to grow the mask by.
        # A higher number will make more and larger masks, ultimately limited by the reference box size.
        mask_growth_steps = {self.mask_growth_steps}
        
        # Maximum threshold for a pixel to be considered off-white.
        # The median color along the edge of a mask may not be pure white,
        # so to prevent slight off-white tones on a pure-white background,
        # anything lighter than this threshold value will be rounded up to pure white.
        off_white_max_threshold = {self.off_white_max_threshold}
        
        # Minimum improvement in standard deviation of the mask to continue shrinking it.
        # The standard deviation refers to the variation is color along the edge of a mask.
        # A low variation means that the mask sits in a solid color,
        # which means it doesn't intersect any text or other objects.
        # Setting a higher value here requires a higher improvement to consider a smaller mask,
        # to give a preference to larger masks.
        mask_improvement_threshold = {self.mask_improvement_threshold}
        
        
        # Whether to use the fast mask selection algorithm.
        # When true, the mask selection algorithm will pick the first perfect mask, if one is found early.
        # This is faster, but may not find the best mask, if a slightly bigger one would have been better.
        mask_selection_fast = {self.mask_selection_fast}
        
        # The maximum standard deviation of a mask to consider.
        # A high value here means a higher tolerance for the mask intersecting text or other objects,
        # which isn't a good mask, as it will require inpainting anyway.
        # Setting this to 0 will only allow perfect masks, which is recommended for very high resolution images.
        mask_max_standard_deviation = {self.mask_max_standard_deviation}
        
        # Color to use for the debug mask. [CLI: This is a tuple of RGBA values.]
        debug_mask_color = {','.join(map(str, self.debug_mask_color))}
        
        """
        masker_conf = cu.ConfigUpdater()
        masker_conf.read_string(multi_left_strip(format_for_version(config_str, gui_mode)))
        masker_section = masker_conf["Masker"]
        config_updater[add_after_section].add_after.space(2).section(masker_section.detach())

    def import_from_conf(self, config_updater: cu.ConfigUpdater) -> None:
        """
        Read the config from the config updater object.

        :param config_updater: An existing config updater object.
        """
        section = "Masker"
        if not config_updater.has_section(section):
            logger.info(f"No {section} section found in the profile, using defaults.")
            return

        try_to_load(self, config_updater, section, int | GreaterZero, "mask_growth_step_pixels")
        try_to_load(self, config_updater, section, int, "mask_growth_steps")
        try_to_load(self, config_updater, section, int, "off_white_max_threshold")
        try_to_load(self, config_updater, section, float, "mask_improvement_threshold")
        try_to_load(self, config_updater, section, bool, "mask_selection_fast")
        try_to_load(self, config_updater, section, float, "mask_max_standard_deviation")
        try:
            color_tuple: tuple[int, ...] = tuple(
                int(x) for x in config_updater["Masker"]["debug_mask_color"].value.split(",")
            )
            if len(color_tuple) != 4:
                raise ValueError(
                    f"Invalid color tuple length. Expected 4: 'Red, Green, Blue, Alpha' got {len(color_tuple)}"
                )
            color_tuple: tuple[int, int, int, int]
            self.debug_mask_color = color_tuple
        except (cu.NoOptionError, ValueError):
            pass

    def fix(self) -> None:
        """
        Keep the numbers greater or equal to zero.
        For numbers, ensure the range 0-255.
        """
        self.mask_growth_steps = max(0, self.mask_growth_steps)
        self.off_white_max_threshold = max(0, min(255, self.off_white_max_threshold))
        if self.mask_improvement_threshold < 0:
            self.mask_improvement_threshold = 0
        if self.mask_max_standard_deviation < 0:
            self.mask_max_standard_deviation = 0
        # We already ensured that it's a tuple of 4 ints.
        # noinspection PyTypeChecker
        self.debug_mask_color = tuple(max(0, min(255, x)) for x in self.debug_mask_color)


@define
class DenoiserConfig:
    denoising_enabled: bool = True
    noise_min_standard_deviation: float = 0.25
    noise_outline_size: int = 5
    noise_fade_radius: int = 1
    colored_images: bool = False
    filter_strength: int = 10
    color_filter_strength: int = 10
    template_window_size: int = 7
    search_window_size: int = 21

    def export_to_conf(
        self, config_updater: cu.ConfigUpdater, add_after_section: str, gui_mode: bool = False
    ) -> None:
        """
        Write the config to the config updater object.

        :param config_updater: An existing config updater object.
        :param add_after_section: The section to add the new section after.
        :param gui_mode: Whether to format the config for the GUI.
        """
        config_str = f"""\
        [Denoiser]
        
        # When a bit of text is hard to mask off, the cleaning step likely had to choose a
        # small mask, which leaves a lot of jpeg-artifacts behind, if they were around the text.
        
        # This means that the color of pixels around the edge of a mask isn't uniform,
        # which is quantified as a standard deviation. The denoiser can target masks with
        # a minimum standard deviation and denoise the area right around the mask.
        # This preserves details in the rest of the image, but removes artifacts right around where
        # the text used to be.
        
        # Since this is an optional step and may even be superfluous for high-resolution images that 
        # don't suffer from jpeg-artifacts, it can be disabled here.
        # Set to False to disable denoising.
        denoising_enabled = {self.denoising_enabled}
        
        # The minimum standard deviation of colors around the edge of a given mask
        # to perform denoising on the region around the mask.
        noise_min_standard_deviation = {self.noise_min_standard_deviation}
        
        # The thickness of an outline to denoise around a mask.
        noise_outline_size = {self.noise_outline_size}
        
        # Fade the edges of the denoised cover mask by this amount to smoothly blend
        # the denoised parts into the rest of the image.
        noise_fade_radius = {self.noise_fade_radius}
        
        # Set to True to support colored images.
        colored_images = {self.colored_images}
        
        # How strongly to denoise the image. Higher values mean more denoising.
        filter_strength = {self.filter_strength}
        
        # How strongly to denoise the color channels, as opposed to lightness.
        # Higher values mean more denoising.
        color_filter_strength = {self.color_filter_strength}
        
        # Size in pixels of the template patch that is used to compute weights. Should be odd.
        template_window_size = {self.template_window_size}
        
        # Size in pixels of the window that is used to compute weighted average for given pixel. Should be odd.
        search_window_size = {self.search_window_size}
        
        """
        denoiser_conf = cu.ConfigUpdater()
        denoiser_conf.read_string(multi_left_strip(format_for_version(config_str, gui_mode)))
        denoiser_section = denoiser_conf["Denoiser"]
        config_updater[add_after_section].add_after.space(2).section(denoiser_section.detach())

    def import_from_conf(self, config_updater: cu.ConfigUpdater) -> None:
        """
        Read the config from the config updater object.

        :param config_updater: An existing config updater object.
        """
        section = "Denoiser"
        if not config_updater.has_section(section):
            logger.info(f"No {section} section found in the profile, using defaults.")
            return

        try_to_load(self, config_updater, section, bool, "denoising_enabled")
        try_to_load(self, config_updater, section, float, "noise_min_standard_deviation")
        try_to_load(self, config_updater, section, int, "noise_outline_size")
        try_to_load(self, config_updater, section, int, "noise_fade_radius")
        try_to_load(self, config_updater, section, bool, "colored_images")
        try_to_load(self, config_updater, section, int, "filter_strength")
        try_to_load(self, config_updater, section, int, "color_filter_strength")
        try_to_load(self, config_updater, section, int, "template_window_size")
        try_to_load(self, config_updater, section, int, "search_window_size")

    def fix(self) -> None:
        if self.noise_min_standard_deviation < 0:
            self.noise_min_standard_deviation = 0
        if self.noise_outline_size < 0:
            self.noise_outline_size = 0
        if self.noise_fade_radius < 0:
            self.noise_fade_radius = 0
        if self.filter_strength < 0:
            self.filter_strength = 0
        if self.color_filter_strength < 0:
            self.color_filter_strength = 0
        if self.template_window_size < 0:
            self.template_window_size = 0
        if self.search_window_size < 0:
            self.search_window_size = 0


@define
class Profile:
    """
    A profile is a collection of settings that can be saved and loaded from disk.
    """

    general: GeneralConfig = field(factory=GeneralConfig)
    text_detector: TextDetectorConfig = field(factory=TextDetectorConfig)
    preprocessor: PreprocessorConfig = field(factory=PreprocessorConfig)
    masker: MaskerConfig = field(factory=MaskerConfig)
    denoiser: DenoiserConfig = field(factory=DenoiserConfig)

    def bundle_config(self, gui_mode: bool = False) -> cu.ConfigUpdater:
        """
        Bundle the config into a ConfigUpdater object.

        :param gui_mode: When true, strips out the CLI-specific options and keeps the GUI-specific ones,
            and vice versa.
        """
        config_updater = cu.ConfigUpdater()
        self.general.export_to_conf(config_updater, gui_mode=gui_mode)
        self.text_detector.export_to_conf(config_updater, "General", gui_mode=gui_mode)
        self.preprocessor.export_to_conf(config_updater, "TextDetector", gui_mode=gui_mode)
        self.masker.export_to_conf(config_updater, "Preprocessor", gui_mode=gui_mode)
        self.denoiser.export_to_conf(config_updater, "Masker", gui_mode=gui_mode)
        return config_updater

    def hash_current_values(self) -> int:
        """
        Hash the current values of the profile.
        This isn't implementing __hash__ because these objects are mutable.
        This is merely useful to tell when the profile has changed.

        :return: A hash of the current values of the profile.
        """
        return hash(str(self.bundle_config()))

    def write(self, path: Path) -> bool:
        """
        Write the profile to a file.

        :param path: The path to write the profile to.
        :return: True if the profile was written successfully, False otherwise.
        """
        logger.debug("Writing profile to disk...")
        try:
            with open(path, "w", encoding="utf-8") as file:
                self.bundle_config().write(file)
            return True
        except Exception as e:
            logger.error(f"Failed to write profile to {path}:\n{e}")
            return False

    @classmethod
    def load(cls, path: Path) -> "Profile":
        """
        Load a profile from a config file.
        """
        logger.debug(f"Loading profile {path} from disk...")
        config = cu.ConfigUpdater()
        try:
            config.read(path, encoding="utf-8")
            profile = cls()
            profile.general.import_from_conf(config)
            profile.text_detector.import_from_conf(config)
            profile.preprocessor.import_from_conf(config)
            profile.masker.import_from_conf(config)
            profile.denoiser.import_from_conf(config)
            profile.fix()
        except Exception as e:
            logger.error(f"Failed to load profile from {path}:\n{e}")
            profile = cls()
            print("Failed to load profile, using default profile.")
        return profile

    def get(self, section: str, option: str) -> Any:
        """
        Get the value of a config option.
        """
        return getattr(getattr(self, section), option)

    def set(self, section: str, option: str, value: Any) -> None:
        """
        Set the value of a config option.
        """
        # Check that the attribute exists.
        if not hasattr(self, section):
            raise AttributeError(f"No such section: {section}")
        if not hasattr(getattr(self, section), option):
            raise AttributeError(f"No such option: {option}")
        setattr(getattr(self, section), option, value)

    def fix(self) -> None:
        """
        Correct any invalid values in the profile.
        """
        self.general.fix()
        self.text_detector.fix()
        self.preprocessor.fix()
        self.masker.fix()
        self.denoiser.fix()


@define
class Config:
    """
    The main configuration class.
    These setting are saved to disk.

    saved_profiles: Saved profiles contains a mapping from profile name to profile path, where it is saved.
    profile_editor: The program to use to edit the profile files. When blank, the default editor is used.
    """

    current_profile: Profile = field(factory=Profile)
    default_profile: str | None = None
    saved_profiles: dict[str, Path] = field(factory=dict)
    profile_editor: str | None = None
    cache_dir: Path | None = None
    default_torch_model_path: Path | None = None  # CUDA
    default_cv2_model_path: Path | None = None  # CPU
    gui_theme: str | None = None

    def show(self) -> None:
        """
        Print the current configuration to the console.
        """
        print("Current Configuration:\n")
        print(
            f"Default Profile: {self.default_profile if self.default_profile is not None else 'Built-in'}"
        )
        print("Saved Profiles:")
        if self.saved_profiles:
            for name, path in self.saved_profiles.items():
                print(f"- {name}: {path}")
        else:
            print("(No saved profiles)")
        print("")

        print(
            "Profile Editor:",
            self.profile_editor if self.profile_editor is not None else "System default",
        )
        print(
            "Cache Directory:", self.cache_dir if self.cache_dir is not None else "System default"
        )
        print(
            "Default Torch Model Path:",
            self.default_torch_model_path
            if self.default_torch_model_path is not None
            else "Not downloaded",
        )
        print(
            "Default CV2 Model Path:",
            self.default_cv2_model_path
            if self.default_cv2_model_path is not None
            else "Not downloaded",
        )
        print(
            "GUI Theme:",
            self.gui_theme if self.gui_theme is not None else "System default",
        )

        print("\n" + "-" * 20 + "\n")
        print(f"Config file located at: {cli.get_config_path()}")
        print(f"System default cache directory: {cli.get_cache_path()}")

    def get_cache_dir(self) -> Path:
        """
        Get the cache directory.
        Get the system default cache directory if the user has not set a custom cache directory.
        """
        if self.cache_dir is None:
            return cli.get_cache_path()
        return self.cache_dir

    def get_model_cache_dir(self) -> Path:
        """
        Get the cache directory for models.
        """
        path = self.get_cache_dir() / "model"
        path.mkdir(parents=True, exist_ok=True)
        return path

    def get_cleaner_cache_dir(self) -> Path:
        """
        Get the cache directory for cleaner models.
        """
        path = self.get_cache_dir() / "cleaner"
        path.mkdir(parents=True, exist_ok=True)
        return path

    def add_profile(self, name: str, path: Path) -> None:
        """
        Add a profile to the saved profiles.
        """
        self.saved_profiles[name] = path

    def remove_profile(self, name: str) -> None:
        """
        Remove a profile from the saved profiles.
        """
        if self.default_profile == name:
            self.default_profile = None
        self.saved_profiles.pop(name, None)

    def save(self, config_path: Path | None = None) -> bool:
        """
        Write the configuration to disk.

        :param config_path: The path to write the config to. If None, the default config path is used.
        :return: True if the configuration was written successfully.
        """
        if config_path is None:
            config_path = cli.get_config_path()

        config = cu.ConfigUpdater()
        # Save the dict of saved profiles and the profile editor.
        saved_profiles_str = "\n".join(
            f"{name}={str(path.resolve())}" for name, path in self.saved_profiles.items()
        )

        conf_str = f"""\
        [General]
        
        # The default profile to use. Leave blank to use the built-in default profile.
        default_profile = {none_to_empty(self.default_profile)}
        
        # This is the command used to open the profile files.
        # If blank, the system default editor is used.
        profile_editor = {none_to_empty(self.profile_editor)}
        
        # This is the directory where the program will store temporary files.
        # If blank, the default directory is used.
        cache_dir = {none_to_empty(self.cache_dir)}
        
        # This is the path to the default torch model.
        # It is downloaded automatically if blank.
        default_torch_model_path = {none_to_empty(self.default_torch_model_path)}
        
        # This is the path to the default cv2 model.
        # It is downloaded automatically if blank.
        default_cv2_model_path = {none_to_empty(self.default_cv2_model_path)}
        
        # This is the theme to use for the GUI.
        # If blank, the system default theme is used.
        # Built-in themes are Breeze and Breeze Dark
        gui_theme = {none_to_empty(self.gui_theme)}
        
        
        [Saved Profiles]
        {saved_profiles_str}
        """
        config.read_string(multi_left_strip(conf_str))
        # Save the config.
        try:
            with open(config_path, "w", encoding="utf-8") as config_file:
                config.write(config_file)
        except OSError as e:
            logger.error(f"Failed to write configuration to {config_path}:\n{e}")
            return False
        return True

    @classmethod
    def from_config_updater(cls, conf_updater: cu.ConfigUpdater) -> "Config":
        """
        Create a Config object from a ConfigUpdater object.
        """

        config = cls()
        section = "General"
        if not conf_updater.has_section(section):
            print(
                f"Failed to load the {section} section from the config file. Proceeding with defaults."
            )
            return config

        try_to_load(config, conf_updater, section, str | None, "default_profile")
        try_to_load(config, conf_updater, section, str | None, "profile_editor")
        try_to_load(config, conf_updater, section, Path | None, "cache_dir")
        config.saved_profiles = {
            k: Path(v.value) for k, v in conf_updater["Saved Profiles"].items()
        }
        try_to_load(config, conf_updater, section, Path | None, "default_torch_model_path")
        try_to_load(config, conf_updater, section, Path | None, "default_cv2_model_path")
        try_to_load(config, conf_updater, section, str | None, "gui_theme")

        # If the default profile isn't in the saved profiles, clear it.
        if (
            config.default_profile is not None
            and config.default_profile not in config.saved_profiles
        ):
            logger.error(
                f"Default profile {config.default_profile} not found in saved profiles. Clearing entry."
            )
            config.default_profile = None

        if config.default_torch_model_path is not None:
            # Verify that the model exists and has the correct hash.
            if not config.default_torch_model_path.is_file():
                print("Configured torch model does not exist. Clearing entry.")
                config.default_torch_model_path = None
            elif not md.check_hash(config.default_torch_model_path, md.TORCH_SHA256):
                print("Configured torch model has the wrong hash. Clearing entry.")
                config.default_torch_model_path = None
        if config.default_cv2_model_path is not None:
            # Verify that the model exists and has the correct hash.
            if not config.default_cv2_model_path.is_file():
                print("Configured cv2 model does not exist. Clearing entry.")
                config.default_cv2_model_path = None
            elif not md.check_hash(config.default_cv2_model_path, md.CV2_SHA256):
                print("Configured cv2 model has the wrong hash. Clearing entry.")
                config.default_cv2_model_path = None

        return config

    def load_profile(self, profile_name: str | None = None) -> tuple[bool, Exception | None]:
        """
        Load a profile from disk, if a name is given.
        First search if the profile is saved, otherwise treat it like a path.
        For None, attempt to load the default profile.

        Special case: Reserve the name "builtin" and "none" (case insensitive) to load the built-in default profile.

        :param profile_name: Name or path of the profile to load.
        :return: True if the profile was loaded successfully.
        """
        logger.debug(f"Loading profile {profile_name!r}...")
        # If no override is given, use the default profile.
        if profile_name is None:
            profile_name = self.default_profile

        # If the default profile is not set, use the builtin default profile.
        if profile_name is None or profile_name.lower() in RESERVED_PROFILE_NAMES:
            logger.debug("Loading builtin default profile")
            self.current_profile = Profile()
            return True, None

        found_profile = cli.closest_match(profile_name, list(self.saved_profiles.keys()))
        if found_profile is not None:
            profile_path = self.saved_profiles[found_profile]
            logger.debug(f"Loading profile {found_profile} from {profile_path}")
        else:
            profile_path = Path(profile_name)
            logger.debug(f"Loading profile from {profile_path}")

        # Try to load the profile from the path.
        try:
            self.current_profile = Profile.load(profile_path)
        except Exception as e:
            logger.error(f"Failed to load profile from {profile_path}:\n{e}")
            self.current_profile = Profile()
            return False, e

        return True, None

    def get_model_path(self, cuda: bool) -> Path:
        """
        Get the path to the default model.
        Check the current profile first. If it is None or the file does not exist,
        return the default model path from the config.
        If it is None, download the model first.

        :param cuda: When true, prefer the torch model.
        """
        if self.current_profile.text_detector.model_path is not None:
            model_path = Path(self.current_profile.text_detector.model_path)
            if model_path.exists():
                return model_path
            else:
                print(
                    f"Model path {model_path} selected in your current profile does not exist. Using default model."
                )

        # Models are downloaded to the folder: cache directory/models
        cache_dir = self.get_model_cache_dir()
        if cuda:
            if self.default_torch_model_path is None:
                self.default_torch_model_path = md.download_torch_model(cache_dir)
                if self.default_torch_model_path is None:
                    print("Failed to find torch model.")
                    sys.exit(1)
                self.save()
            return self.default_torch_model_path
        else:
            if self.default_cv2_model_path is None:
                self.default_cv2_model_path = md.download_cv2_model(cache_dir)
                if self.default_cv2_model_path is None:
                    print("Failed to find cv2 model.")
                    sys.exit(1)
                self.save()
            return self.default_cv2_model_path


def load_config() -> Config:
    """
    Try to load the config from disk, otherwise create a new one.
    :return:
    """
    config_path = cli.get_config_path()

    if config_path.is_file():
        conf_updater = cu.ConfigUpdater()
        try:
            conf_updater.read(config_path, encoding="utf-8")
        except OSError as e:
            logger.error(f"Failed to read configuration from {config_path}:\n{e}")
            if cli.get_confirmation("Do you want to create a new configuration file?"):
                config = Config()
                if not config.save(config_path):
                    print("\n\nProceeding with default configuration.")
                    return config
            else:
                print("Aborting.")
                sys.exit(1)
        # Parse the config from the ConfigUpdater.
        try:
            config = Config.from_config_updater(conf_updater)
        except cu.ParsingError as e:
            # When configupdater fails to parse the config, it raises a KeyError of the form:
            # ('No option `thing` found', {'key': 'thing'})
            if isinstance(e.args, tuple):
                e = e.args[0]
            logger.error(f"Failed to parse configuration from {config_path}:\n{e}")
            if cli.get_confirmation("Do you want to create a new configuration file?"):
                config = Config()
                if not config.save(config_path):
                    print("\n\nProceeding with default configuration.")
                    return config
            else:
                print("Aborting.")
                sys.exit(1)

    else:
        print("No config file found, creating a new one.")
        config = Config()
        config.save(config_path)

    return config


def try_to_load(
    obj: object, conf_updater: cu.ConfigUpdater, section: str, attr_type, attr_name: str
):
    """
    Try to load the attribute from the ConfigUpdater, validating the type.
    The key and the attribute name need to match.

    Supports attr_type:
    bool, int, float, str, Path, str | None, Path | None

    Union types with None will return none if the bare string is empty.

    :param obj: The object to assign it to.
    :param attr_name: The name of the attribute to load.
    :param section: The config section name.
    :param attr_type: The type to construct.
    :param conf_updater: The ConfigUpdater object to load from.
    """
    try:
        conf_data = conf_updater.get(section, attr_name).value
    except cu.NoOptionError as e:
        print(f"Option {attr_name} not found in {section}. Using default.")
        logger.debug(e)
        return

    # Handle bool specially.
    if attr_type == bool:
        if conf_data.lower() in ("1", "t", "true", "y", "yes", "on"):
            attr_value = True
        elif conf_data.lower() in ("0", "f", "false", "n", "no", "off"):
            attr_value = False
        else:
            print(
                f"Option {attr_name} in section {section} should be True or False, not '{conf_data}'"
            )
            return
    elif attr_type == str:
        attr_value = conf_data
    elif attr_type == str | None:
        if conf_data == "":
            attr_value = None
        else:
            attr_value = conf_data
    elif attr_type == int:
        try:
            attr_value = int(conf_data)
        except ValueError as e:
            print(
                f"Option {attr_name} in section {section} should be an integer.\n"
                f"Failed to parse '{conf_data}': {e}"
            )
            return
    elif attr_type == float:
        try:
            attr_value = float(conf_data)
        except ValueError as e:
            print(
                f"Option {attr_name} in section {section} should be floating point number.\n"
                f"Failed to parse '{conf_data}': {e}"
            )
            return
    elif attr_type == int | GreaterZero:
        # GreaterZero is just a signifier to check for positive numbers.
        try:
            attr_value = int(conf_data)
        except ValueError as e:
            print(
                f"Option {attr_name} in section {section} should be an integer.\n"
                f"Failed to parse '{conf_data}': {e}"
            )
            return
        if attr_value <= 0:
            print(
                f"Option {attr_name} in section {section} should be greater than zero. Limiting to 1."
            )
            attr_value = 1

    elif attr_type == float | GreaterZero:
        # GreaterZero is just a signifier to check for positive numbers.
        try:
            attr_value = float(conf_data)
        except ValueError as e:
            print(
                f"Option {attr_name} in section {section} should be floating point number.\n"
                f"Failed to parse '{conf_data}': {e}"
            )
            return
        if attr_value <= 0:
            print(
                f"Option {attr_name} in section {section} should be greater than zero. Setting to 1."
            )
            attr_value = 0.01

    elif attr_type == Path:
        try:
            attr_value = Path(conf_data)
        except ValueError as e:
            print(
                f"Option {attr_name} in section {section} should be file path.\n"
                f"Failed to parse '{conf_data}': {e}"
            )
            return
    elif attr_type == Path | None:
        if conf_data == "":
            attr_value = None
        else:
            try:
                attr_value = Path(conf_data)
            except ValueError as e:
                print(
                    f"Option {attr_name} in section {section} should be file path.\n"
                    f"Failed to parse '{conf_data}': {e}"
                )
                return
    else:
        logger.error(f"Unsupported attribute type {attr_type}.")
        return

    # Assign value.
    obj.__setattr__(attr_name, attr_value)


def none_to_empty(value: str | None) -> str:
    return "" if value is None else value


def multi_left_strip(string: str) -> str:
    """
    Strip all leading whitespace from all lines in a string.

    :param string: The string to strip.
    :return: The stripped string.
    """
    stripped_lines = (line.lstrip() for line in string.splitlines())
    return "\n".join(stripped_lines)


def format_for_version(conf_string: str, gui_mode: bool) -> str:
    """
    Format the config string for the current version of the program.

    In CLI mode, remove all GUI-only options, and vice versa.
    The comments that need to be replaces are marked as follows:
    [CLI: ...] and [GUI: ...]

    :param conf_string: The config string to format.
    :param gui_mode: Whether to format for GUI mode.
    :return: The formatted config string.
    """
    if gui_mode:
        conf_string = re.sub(r"^\s*\[CLI: (.*?)]\s*$", "", conf_string, flags=re.MULTILINE)
        conf_string = re.sub(r"^\s*\[GUI: (.*?)]\s*$", r"\1", conf_string, flags=re.MULTILINE)
        conf_string = re.sub(r"\[CLI: (.*?)]", "", conf_string)
        conf_string = re.sub(r"\[GUI: (.*?)]", r"\1", conf_string)
    else:
        conf_string = re.sub(r"^\s*\[GUI: (.*?)]\s*$", "", conf_string, flags=re.MULTILINE)
        conf_string = re.sub(r"^\s*\[CLI: (.*?)]\s*$", r"\1", conf_string, flags=re.MULTILINE)
        conf_string = re.sub(r"\[GUI: (.*?)]", "", conf_string)
        conf_string = re.sub(r"\[CLI: (.*?)]", r"\1", conf_string)

    return conf_string



from pathlib import Path

import PySide6.QtCore as Qc
import PySide6.QtGui as Qg
import PySide6.QtWidgets as Qw
from loguru import logger


ZOOM_TICK_FACTOR = 1.25


class ImageViewer(Qw.QGraphicsView):
    mouseMoved = Qc.Signal(int, int)
    image_dimensions: tuple[int, int] | None
    image_item: Qw.QGraphicsPixmapItem | None

    def __init__(self, parent=None) -> None:
        super(ImageViewer, self).__init__(parent)
        self.setRenderHint(Qg.QPainter.Antialiasing)

        # Scene for QGraphicsView
        self.scene = Qw.QGraphicsScene(self)
        self.setScene(self.scene)

        self.image_item = None
        # Disable the allocation limit for high resolution images.
        Qg.QImageReader.setAllocationLimit(0)

        self.zoom_factor = 1.0
        self.setAlignment(Qc.Qt.AlignCenter)

        self.setMouseTracking(True)

    def pixmap_valid(self) -> bool:
        return self.image_item is not None and self.image_item.pixmap().isNull()

    def set_image(self, image_path: Path = None) -> None:
        if image_path:
            self.scene.clear()
            self.image_item = Qw.QGraphicsPixmapItem()
            self.scene.addItem(self.image_item)
            # Move the image item by -0.5, -0.5 to make it align with the scene's
            # origin (0, 0).
            self.image_item.setOffset(-0.5, -0.5)

            image = Qg.QImage(str(image_path))
            pixmap = Qg.QPixmap.fromImage(image)
            self.image_item.setPixmap(pixmap)
            self.setSceneRect(pixmap.rect())
            dim = self.image_item.pixmap().size()
            self.image_dimensions = (dim.width(), dim.height())

        else:
            # Display "nothing" message when no image is set
            self.scene.clear()
            self.image_dimensions = None
            self.image_item = None

    def wheelEvent(self, event: Qg.QWheelEvent) -> None:
        if Qc.Qt.ControlModifier & event.modifiers():
            # Zoom in/out with Ctrl + mouse wheel
            if event.angleDelta().y() > 0:
                self.zoom_in(wheel=True)
            else:
                self.zoom_out(wheel=True)
        elif Qc.Qt.ShiftModifier & event.modifiers():
            # Horizontal scrolling with Shift + mouse wheel
            self.horizontalScrollBar().setValue(
                self.horizontalScrollBar().value() - event.angleDelta().y()
            )
        else:
            # Default behavior for panning
            super().wheelEvent(event)

    # For the mouse wheel events, we want to zoom slower.
    # Since zooming applies a factor, we can use a root of the factor to
    # achieve half of a zoom step.
    def zoom_in(self, wheel=False) -> None:
        if wheel:
            self.zoom(ZOOM_TICK_FACTOR**0.5)
        else:
            self.zoom(ZOOM_TICK_FACTOR)

    def zoom_out(self, wheel=False) -> None:
        if wheel:
            self.zoom(1 / (ZOOM_TICK_FACTOR**0.5))
        else:
            self.zoom(1 / ZOOM_TICK_FACTOR)

    def zoom_reset(self) -> None:
        self.zoom_factor = 1.0
        self.zoom(1)

    def zoom_fit(self) -> None:
        if self.image_dimensions is None:
            return
        width, height = self.image_dimensions
        view_width, view_height = self.viewport().width(), self.viewport().height()
        width_factor = view_width / width
        height_factor = view_height / height
        self.zoom_factor = min(width_factor, height_factor)
        logger.debug(
            f"Zoom to fit with factor: {self.zoom_factor:.2%} for view size {view_width}x{view_height} and image size {width}x{height}"
        )
        self.zoom(1)

    def image_position(self, pos) -> Qc.QPoint:
        return self.mapToScene(pos).toPoint()

    def mouseMoveEvent(self, event) -> None:
        if self.pixmap_valid():
            return
        # Call base class implementation for standard behavior
        super().mouseMoveEvent(event)
        # Emit the mouse position always
        image_pos = self.image_position(event.pos())
        self.mouseMoved.emit(image_pos.x(), image_pos.y())
        self.viewport().update()  # Request redraw for the pixel highlight

    def drawForeground(self, painter, rect) -> None:
        if not self.pixmap_valid() and self.zoom_factor > 5:
            view_pos = self.mapFromGlobal(Qg.QCursor.pos())
            image_pos = self.mapToScene(view_pos).toPoint()

            # Check if the cursor is inside the pixmap area
            if self.image_item.pixmap().rect().contains(image_pos):
                # Decide the color for the square
                pixel_color = self.image_item.pixmap().toImage().pixelColor(image_pos)
                avg = (pixel_color.red() + pixel_color.green() + pixel_color.blue()) / 3
                square_color = Qc.Qt.white if avg < 128 else Qc.Qt.black

                painter.setPen(Qg.QPen(square_color, 1 / self.zoom_factor))
                rect = Qc.QRectF(image_pos.x() - 0.5, image_pos.y() - 0.5, 1, 1)
                scaled_rect = rect.adjusted(
                    -0.5 / self.zoom_factor,
                    -0.5 / self.zoom_factor,
                    0.5 / self.zoom_factor,
                    0.5 / self.zoom_factor,
                )
                painter.drawRect(scaled_rect)

    def zoom(self, factor) -> None:
        """
        Zoom the image by the given factor.
        The image may not exceed a scale of 100x or have both width and height be smaller
        than half of the viewport's width and height.

        :param factor: The factor to multiply the current zoom factor by.
        """
        proposed_zoom_factor = self.zoom_factor * factor
        proposed_zoom_factor = min(proposed_zoom_factor, 100)

        current_width, current_height = (
            self.image_item.pixmap().width(),
            self.image_item.pixmap().height(),
        )
        proposed_width, proposed_height = (
            current_width * proposed_zoom_factor,
            current_height * proposed_zoom_factor,
        )
        view_width, view_height = self.viewport().width(), self.viewport().height()

        # Don't zoom out further if it's getting too smol.
        if proposed_width < view_width / 2 and proposed_height < view_height / 2 and factor < 1:
            return

        self.zoom_factor = proposed_zoom_factor

        if self.zoom_factor >= 2:
            self.setRenderHint(Qg.QPainter.SmoothPixmapTransform, False)
            self.image_item.setTransformationMode(Qc.Qt.FastTransformation)
        else:
            self.setRenderHint(Qg.QPainter.SmoothPixmapTransform, True)
            self.image_item.setTransformationMode(Qc.Qt.SmoothTransformation)
        self.setTransform(Qg.QTransform().scale(self.zoom_factor, self.zoom_factor))

    def reset_zoom(self) -> None:
        self.zoom_factor = 1
        self.setTransform(Qg.QTransform().scale(self.zoom_factor, self.zoom_factor))



import datetime

import PySide6
import PySide6.QtGui as Qg
import PySide6.QtWidgets as Qw
from PySide6.QtCore import Qt

import pcleaner.gui.license_driver as ld
from pcleaner import __version__
from pcleaner.gui.ui_generated_files.ui_About import Ui_About


class AboutWidget(Qw.QWidget, Ui_About):
    """
    Displays the about information.
    """

    def __init__(
        self,
        parent: Qw.QWidget | None = None,
    ):
        """
        Initialize the widget.

        :param parent: The parent widget.
        """
        Qw.QWidget.__init__(self, parent)
        self.setupUi(self)
        self.setWindowFlag(Qt.Window)
        self.setWindowIcon(Qg.QIcon(":/logo-tiny.png"))

        self.label_license.linkActivated.connect(self.open_license)

        copyright_str = f"© 2023"
        if until := datetime.datetime.now().year != 2023:
            copyright_str += f"–{until}"

        self.label_copyright.setText(copyright_str)

        self.label_version.setText(__version__)
        self.label_toolkit.setText(f"PySide (Qt) {PySide6.__version__}")

        self.label_logo.setPixmap(
            Qg.QPixmap(":/logo.png").scaledToWidth(200, mode=Qt.SmoothTransformation)
        )

    def open_license(self) -> None:
        """
        Open the license dialog.
        """
        license_widget = ld.LicenseDialog(self)
        license_widget.show()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'SetupGreeter.ui'
##
## Created by: Qt User Interface Compiler version 6.5.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractButton, QApplication, QDialog, QDialogButtonBox,
    QFormLayout, QLabel, QSizePolicy, QSpacerItem,
    QVBoxLayout, QWidget)

class Ui_SetupGreeter(object):
    def setupUi(self, SetupGreeter):
        if not SetupGreeter.objectName():
            SetupGreeter.setObjectName(u"SetupGreeter")
        SetupGreeter.resize(720, 382)
        SetupGreeter.setMinimumSize(QSize(600, 0))
        SetupGreeter.setModal(True)
        self.verticalLayout = QVBoxLayout(SetupGreeter)
        self.verticalLayout.setSpacing(20)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(-1, 20, -1, -1)
        self.label_2 = QLabel(SetupGreeter)
        self.label_2.setObjectName(u"label_2")
        self.label_2.setTextFormat(Qt.MarkdownText)

        self.verticalLayout.addWidget(self.label_2, 0, Qt.AlignHCenter)

        self.label = QLabel(SetupGreeter)
        self.label.setObjectName(u"label")
        self.label.setTextFormat(Qt.PlainText)
        self.label.setWordWrap(True)
        self.label.setOpenExternalLinks(True)
        self.label.setTextInteractionFlags(Qt.NoTextInteraction)

        self.verticalLayout.addWidget(self.label)

        self.formLayout = QFormLayout()
        self.formLayout.setObjectName(u"formLayout")
        self.formLayout.setFieldGrowthPolicy(QFormLayout.ExpandingFieldsGrow)
        self.formLayout.setRowWrapPolicy(QFormLayout.WrapLongRows)
        self.label_9 = QLabel(SetupGreeter)
        self.label_9.setObjectName(u"label_9")

        self.formLayout.setWidget(0, QFormLayout.LabelRole, self.label_9)

        self.label_3 = QLabel(SetupGreeter)
        self.label_3.setObjectName(u"label_3")
        sizePolicy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_3.sizePolicy().hasHeightForWidth())
        self.label_3.setSizePolicy(sizePolicy)
        self.label_3.setTextFormat(Qt.MarkdownText)
        self.label_3.setWordWrap(True)
        self.label_3.setOpenExternalLinks(True)
        self.label_3.setTextInteractionFlags(Qt.TextBrowserInteraction)

        self.formLayout.setWidget(0, QFormLayout.FieldRole, self.label_3)

        self.label_4 = QLabel(SetupGreeter)
        self.label_4.setObjectName(u"label_4")
        self.label_4.setTextFormat(Qt.PlainText)
        self.label_4.setWordWrap(True)
        self.label_4.setOpenExternalLinks(True)
        self.label_4.setTextInteractionFlags(Qt.NoTextInteraction)

        self.formLayout.setWidget(1, QFormLayout.LabelRole, self.label_4)

        self.label_text_detector_path = QLabel(SetupGreeter)
        self.label_text_detector_path.setObjectName(u"label_text_detector_path")
        self.label_text_detector_path.setTextInteractionFlags(Qt.TextSelectableByKeyboard|Qt.TextSelectableByMouse)

        self.formLayout.setWidget(1, QFormLayout.FieldRole, self.label_text_detector_path)

        self.label_10 = QLabel(SetupGreeter)
        self.label_10.setObjectName(u"label_10")
        self.label_10.setWordWrap(True)

        self.formLayout.setWidget(3, QFormLayout.LabelRole, self.label_10)

        self.label_5 = QLabel(SetupGreeter)
        self.label_5.setObjectName(u"label_5")
        sizePolicy1 = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.label_5.sizePolicy().hasHeightForWidth())
        self.label_5.setSizePolicy(sizePolicy1)
        self.label_5.setTextFormat(Qt.MarkdownText)
        self.label_5.setWordWrap(True)
        self.label_5.setOpenExternalLinks(True)
        self.label_5.setTextInteractionFlags(Qt.TextBrowserInteraction)

        self.formLayout.setWidget(3, QFormLayout.FieldRole, self.label_5)

        self.label_7 = QLabel(SetupGreeter)
        self.label_7.setObjectName(u"label_7")
        self.label_7.setTextFormat(Qt.PlainText)
        self.label_7.setWordWrap(True)
        self.label_7.setOpenExternalLinks(True)
        self.label_7.setTextInteractionFlags(Qt.NoTextInteraction)

        self.formLayout.setWidget(4, QFormLayout.LabelRole, self.label_7)

        self.label_ocr_path = QLabel(SetupGreeter)
        self.label_ocr_path.setObjectName(u"label_ocr_path")
        self.label_ocr_path.setTextInteractionFlags(Qt.TextSelectableByKeyboard|Qt.TextSelectableByMouse)

        self.formLayout.setWidget(4, QFormLayout.FieldRole, self.label_ocr_path)

        self.verticalSpacer_2 = QSpacerItem(20, 20, QSizePolicy.Minimum, QSizePolicy.Fixed)

        self.formLayout.setItem(2, QFormLayout.LabelRole, self.verticalSpacer_2)


        self.verticalLayout.addLayout(self.formLayout)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.verticalLayout.addItem(self.verticalSpacer)

        self.buttonBox = QDialogButtonBox(SetupGreeter)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Abort|QDialogButtonBox.Ok)

        self.verticalLayout.addWidget(self.buttonBox)


        self.retranslateUi(SetupGreeter)
        self.buttonBox.accepted.connect(SetupGreeter.accept)
        self.buttonBox.rejected.connect(SetupGreeter.reject)

        QMetaObject.connectSlotsByName(SetupGreeter)
    # setupUi

    def retranslateUi(self, SetupGreeter):
        SetupGreeter.setWindowTitle(QCoreApplication.translate("SetupGreeter", u"Downloading Machine Learning Models", None))
        self.label_2.setText(QCoreApplication.translate("SetupGreeter", u"## Setting up Panel Cleaner", None))
        self.label.setText(QCoreApplication.translate("SetupGreeter", u"Before Panel Cleaner can begin, it requires the following machine learning models. Press OK to allow Panel Cleaner to automatically download these files to get started (Internet connection required, total space approx. 524 MiB)", None))
        self.label_9.setText(QCoreApplication.translate("SetupGreeter", u"Text Detection:", None))
        self.label_3.setText(QCoreApplication.translate("SetupGreeter", u"[https://github.com/zyddnys/manga-image-translator/releases](https://github.com/zyddnys/manga-image-translator/releases) \\\n"
"(only the comictextdetector file is needed)", None))
        self.label_4.setText(QCoreApplication.translate("SetupGreeter", u"Saved in:", None))
        self.label_text_detector_path.setText(QCoreApplication.translate("SetupGreeter", u"<text detection path>", None))
        self.label_10.setText(QCoreApplication.translate("SetupGreeter", u"Optical Character Recognition (OCR):", None))
        self.label_5.setText(QCoreApplication.translate("SetupGreeter", u"[https://huggingface.co/kha-white/manga-ocr-base/tree/main](https://huggingface.co/kha-white/manga-ocr-base/tree/main)", None))
        self.label_7.setText(QCoreApplication.translate("SetupGreeter", u"Saved in:", None))
        self.label_ocr_path.setText(QCoreApplication.translate("SetupGreeter", u"<ocr path>", None))
    # retranslateUi




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'ModelDownloader.ui'
##
## Created by: Qt User Interface Compiler version 6.5.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QDialog, QHBoxLayout, QLabel,
    QProgressBar, QSizePolicy, QSpacerItem, QVBoxLayout,
    QWidget)

class Ui_ModelDownloader(object):
    def setupUi(self, ModelDownloader):
        if not ModelDownloader.objectName():
            ModelDownloader.setObjectName(u"ModelDownloader")
        ModelDownloader.resize(720, 209)
        ModelDownloader.setModal(True)
        self.verticalLayout = QVBoxLayout(ModelDownloader)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.label = QLabel(ModelDownloader)
        self.label.setObjectName(u"label")

        self.verticalLayout.addWidget(self.label)

        self.verticalSpacer_2 = QSpacerItem(20, 10, QSizePolicy.Minimum, QSizePolicy.Fixed)

        self.verticalLayout.addItem(self.verticalSpacer_2)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setSpacing(12)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.label_model_name = QLabel(ModelDownloader)
        self.label_model_name.setObjectName(u"label_model_name")

        self.horizontalLayout.addWidget(self.label_model_name)

        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer)

        self.label_model_speed = QLabel(ModelDownloader)
        self.label_model_speed.setObjectName(u"label_model_speed")

        self.horizontalLayout.addWidget(self.label_model_speed)

        self.label_model_size = QLabel(ModelDownloader)
        self.label_model_size.setObjectName(u"label_model_size")

        self.horizontalLayout.addWidget(self.label_model_size)


        self.verticalLayout.addLayout(self.horizontalLayout)

        self.progressBar_model = QProgressBar(ModelDownloader)
        self.progressBar_model.setObjectName(u"progressBar_model")
        self.progressBar_model.setValue(0)

        self.verticalLayout.addWidget(self.progressBar_model)

        self.verticalSpacer = QSpacerItem(20, 20, QSizePolicy.Minimum, QSizePolicy.Fixed)

        self.verticalLayout.addItem(self.verticalSpacer)

        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setSpacing(12)
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.label_2 = QLabel(ModelDownloader)
        self.label_2.setObjectName(u"label_2")

        self.horizontalLayout_2.addWidget(self.label_2)

        self.label_ocr_file_name = QLabel(ModelDownloader)
        self.label_ocr_file_name.setObjectName(u"label_ocr_file_name")

        self.horizontalLayout_2.addWidget(self.label_ocr_file_name)

        self.horizontalSpacer_2 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout_2.addItem(self.horizontalSpacer_2)

        self.label_ocr_speed = QLabel(ModelDownloader)
        self.label_ocr_speed.setObjectName(u"label_ocr_speed")

        self.horizontalLayout_2.addWidget(self.label_ocr_speed)

        self.label_ocr_size = QLabel(ModelDownloader)
        self.label_ocr_size.setObjectName(u"label_ocr_size")

        self.horizontalLayout_2.addWidget(self.label_ocr_size)


        self.verticalLayout.addLayout(self.horizontalLayout_2)

        self.progressBar_ocr = QProgressBar(ModelDownloader)
        self.progressBar_ocr.setObjectName(u"progressBar_ocr")
        self.progressBar_ocr.setValue(0)

        self.verticalLayout.addWidget(self.progressBar_ocr)

        self.verticalSpacer_3 = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.verticalLayout.addItem(self.verticalSpacer_3)


        self.retranslateUi(ModelDownloader)

        QMetaObject.connectSlotsByName(ModelDownloader)
    # setupUi

    def retranslateUi(self, ModelDownloader):
        ModelDownloader.setWindowTitle(QCoreApplication.translate("ModelDownloader", u"Downloading Machine Learning Models", None))
        self.label.setText(QCoreApplication.translate("ModelDownloader", u"Model data required to run Panel Cleaner:", None))
        self.label_model_name.setText(QCoreApplication.translate("ModelDownloader", u"<model name>", None))
        self.label_model_speed.setText(QCoreApplication.translate("ModelDownloader", u"<download speed>", None))
        self.label_model_size.setText(QCoreApplication.translate("ModelDownloader", u"<download size/total>", None))
        self.label_2.setText(QCoreApplication.translate("ModelDownloader", u"OCR model:", None))
        self.label_ocr_file_name.setText(QCoreApplication.translate("ModelDownloader", u"<file name>", None))
        self.label_ocr_speed.setText(QCoreApplication.translate("ModelDownloader", u"<download speed>", None))
        self.label_ocr_size.setText(QCoreApplication.translate("ModelDownloader", u"<download size/total>", None))
    # retranslateUi

